# -*- coding: utf-8 -*-

# Python3 compatiability
from __future__ import print_function
import sys
if sys.version_info > (3,):
    xrange = range
    
"""
Deconvolution support for images made with :func:`lsl.imaging.utils.buildGriddedImage`.
"""

import numpy
from aipy.coord import eq2radec, top2azalt
from aipy.fit import RadioFixedBody
from scipy.signal import fftconvolve as convolve

from lsl.sim.vis import buildSimData
from lsl.imaging import utils
from lsl.astro import deg_to_dms, deg_to_hms
from lsl.statistics.robust import std as rStd
from lsl.misc.mathutil import gaussian2d

__version__ = '0.5'
__revision__ = '$Rev: 2457 $'
__all__ = ['clean', 'cleanSources', 'lsq', '__version__', '__revision__', '__all__']


def _interpolateValues(data, peakX, peakY):
    x1 = int(peakX)
    x2 = x1 + 1
    y1 = int(peakY)
    y2 = y1 + 1
    
    x = peakX
    if x < 0:
        raise IndexError("Invalid x")
    y = peakY
    if y < 0:
        raise IndexError("Invalid y")
        
    dataPrime  = data[x1,y1]*(x2-x)*(y2-y)
    dataPrime += data[x2,y1]*(x-x1)*(y2-y)
    dataPrime += data[x1,y2]*(x2-x)*(y-y1)
    dataPrime += data[x2,y2]*(x-x1)*(y-y1)
    dataPrime /= (x2-x1)*(y2-y1)
    return dataPrime


def _fit2DGaussian(data):
    """
    Fit a 2D Gaussian to the provided data.  This function returns a
    five-element tuple of:
    * height
    * center - X and Y
    * sigma - X and Y
    """
    
    from scipy.optimize import leastsq
    
    def gaussian(height, center_x, center_y, width_x, width_y):
        """
        Returns a gaussian function with the given parameters
        
        From:  http://wiki.scipy.org/Cookbook/FittingData
        """
        
        width_x = float(width_x)
        width_y = float(width_y)
        return lambda x,y: height*numpy.exp(
                    -(((center_x-x)/width_x)**2+((center_y-y)/width_y)**2)/2)

    def moments(data):
        """
        Returns (height, x, y, width_x, width_y) the gaussian parameters 
        of a 2D distribution by calculating its moments
        
        From:  http://wiki.scipy.org/Cookbook/FittingData
        """
        
        total = data.sum()
        X, Y = numpy.indices(data.shape)
        x = (X*data).sum()/total
        y = (Y*data).sum()/total
        col = data[:, int(y)]
        width_x = numpy.sqrt(abs((numpy.arange(col.size)-y)**2*col).sum()/col.sum())
        row = data[int(x), :]
        width_y = numpy.sqrt(abs((numpy.arange(row.size)-x)**2*row).sum()/row.sum())
        height = data.max()
        return height, x, y, width_x, width_y

    def fitgaussian(data):
        """
        Returns (height, x, y, width_x, width_y) the gaussian parameters 
        of a 2D distribution found by a fit
        
        From:  http://wiki.scipy.org/Cookbook/FittingData
        """
        
        params = moments(data)
        errorfunction = lambda p: numpy.ravel(gaussian(*p)(*numpy.indices(data.shape)) -
                                data)
        p, success = leastsq(errorfunction, params)
        return p
        
    params = fitgaussian(data)
    fit = gaussian(*params)
    
    #pylab.matshow(data, cmap=pylab.cm.gist_earth_r)
    #pylab.contour(fit(*numpy.indices(data.shape)), cmap=pylab.cm.copper)
    #pylab.show()
    
    return params


def clean(aa, dataDict, aipyImg, imageInput=None, MapSize=80, MapRes=0.50, MapWRes=0.10, pol='xx', chan=None, gain=0.2, maxIter=150, sigma=3.0, verbose=True, plot=False):
    """
    Given a AIPY antenna array instance, a data dictionary, and an AIPY ImgW 
    instance filled with data, return a deconvolved image.  This function 
    uses a CLEAN-like method that computes the array beam for each peak in 
    the flux.  Thus the CLEAN loop becomes:
      1.  Find the peak flux in the residual image
      2.  Compute the systems response to a point source at that location
      3.  Remove the scaled porition of this beam from the residuals
      4.  Go to 1.
    
    CLEAN tuning parameters:
      * gain - CLEAN loop gain (default 0.2)
      * maxIter - Maximum number of iterations (default 150)
      * sigma - Threshold in sigma to stop cleaning (default 3.0)
    """
    
    # Sort out the channels to work on
    if chan is None:
        chan = range(dataDict['freq'].size)
        
    # Get a grid of right ascensions and dec values for the image we are working with
    xyz = aipyImg.get_eq(0.0, aa.lat, center=(MapSize,MapSize))
    RA, dec = eq2radec(xyz)
    RA += aa.sidereal_time()
    RA %= (2*numpy.pi)
    top = aipyImg.get_top(center=(MapSize,MapSize))
    az,alt = top2azalt(top)
    
    # Get the list of baselines to generate visibilites for
    baselines = dataDict['bls'][pol]
    
    # Get the actual image out of the ImgW instance
    if imageInput is None:
        img = aipyImg.image(center=(MapSize,MapSize))
    else:
        img = imageInput*1.0
        
    # Setup the arrays to hold the point sources and the residual.
    cleaned = numpy.zeros_like(img)
    working = numpy.zeros_like(img)
    working += img
    
    # Setup the dictionary that will hold the beams as they are computed
    prevBeam = {}
    
    # Estimate the zenith beam response
    psfSrc = {'z': RadioFixedBody(aa.sidereal_time(), aa.lat, jys=1.0, index=0, epoch=aa.date)}
    psfDict = buildSimData(aa, psfSrc, jd=aa.get_jultime(), pols=[pol,], chan=chan, baselines=baselines, flatResponse=True)
    psf = utils.buildGriddedImage(psfDict, MapSize=MapSize, MapRes=MapRes, MapWRes=MapWRes, chan=chan, pol=pol, verbose=verbose)
    psf = psf.image(center=(MapSize,MapSize))
    psf /= psf.max()
    
    # Fit a Guassian to the zenith beam response and use that for the restore beam
    beamCutout = psf[MapSize/2:3*MapSize/2, MapSize/2:3*MapSize/2]
    beamCutout = numpy.where( beamCutout > 0.0, beamCutout, 0.0 )
    h, cx, cy, sx, sy = _fit2DGaussian( beamCutout )
    gauGen = gaussian2d(1.0, MapSize/2+cx, MapSize/2+cy, sx, sy)
    FWHM = int( round( (sx+sy)/2.0 * 2.0*numpy.sqrt(2.0*numpy.log(2.0)) ) )
    beamClean = psf * 0.0
    for i in xrange(beamClean.shape[0]):
        for j in xrange(beamClean.shape[1]):
            beamClean[i,j] = gauGen(i,j)
    beamClean /= beamClean.sum()
    convMask = xyz.mask[0,:,:]
    
    # Go!
    if plot:
        import pylab
        from matplotlib import pyplot as plt
        
        pylab.ion()
        
    exitStatus = 'iteration limit'
    for i in xrange(maxIter):
        # Find the location of the peak in the flux density
        peak = numpy.where( working == working.max() )
        peakX = peak[0][0]
        peakY = peak[1][0]
        peakV = working[peakX,peakY]
        
        # Optimize the location
        peakParams = _fit2DGaussian(working[peakX-FWHM/2:peakX+FWHM/2+1, peakY-FWHM/2:peakY+FWHM/2+1])
        peakVO = peakParams[0]
        peakXO = peakX - FWHM/2 + peakParams[1]
        peakYO = peakY - FWHM/2 + peakParams[2]
        
        # Quantize to try and keep the computation down without over-simplifiying things
        subpixelationLevel = 5
        peakXO = round(peakXO*subpixelationLevel)/float(subpixelationLevel)
        peakYO = round(peakYO*subpixelationLevel)/float(subpixelationLevel)
        
        # Pixel coordinates to right ascension, dec.
        try:
            peakRA = _interpolateValues(RA, peakXO, peakYO)
        except IndexError:
            peakRA = RA[peakX, peakY]
        try:
            peakDec = _interpolateValues(dec, peakXO, peakYO)
        except IndexError:
            peakDec = dec[peakX, peakY]
            
        # Pixel coordinates to az, el
        try:
            peakAz = _interpolateValues(az, peakXO, peakYO)
        except IndexError:
            peakAz = az[peakX, peakY]
        try:
            peakEl = _interpolateValues(alt, peakX, peakY)
        except IndexError:
            peakEl = alt[peakX, peakY]
            
        if verbose:
            currRA  = deg_to_hms(peakRA * 180/numpy.pi)
            currDec = deg_to_dms(peakDec * 180/numpy.pi)
            currAz  = deg_to_dms(peakAz * 180/numpy.pi)
            currEl  = deg_to_dms(peakEl * 180/numpy.pi)
            
            print("Iteration %i:  Log peak of %.3f at row: %i, column: %i" % (i+1, numpy.log10(peakV), peakX, peakY))
            print("               -> RA: %s, Dec: %s" % (currRA, currDec))
            print("               -> az: %s, el: %s" % (currAz, currEl))
            
        # Check for the exit criteria
        if peakV < 0:
            exitStatus = 'peak value is negative'
            
            break
        
        # Find the beam index and see if we need to compute the beam or not
        beamIndex = (int(peakXO*subpixelationLevel), int(peakYO*subpixelationLevel))
        try:
            beam = prevBeam[beamIndex]
            
        except KeyError:
            if verbose:
                print("               -> Computing beam(s)")
                
            beamSrc = {'Beam': RadioFixedBody(peakRA, peakDec, jys=1.0, index=0, epoch=aa.date)}
            beamDict = buildSimData(aa, beamSrc, jd=aa.get_jultime(), pols=[pol,], chan=chan, baselines=baselines, flatResponse=True)
            beam = utils.buildGriddedImage(beamDict, MapSize=MapSize, MapRes=MapRes, MapWRes=MapWRes, chan=chan, pol=pol, verbose=verbose)
            beam = beam.image(center=(MapSize,MapSize))
            beam /= beam.max()
            print("                  ", beam.mean(), beam.min(), beam.max(), beam.sum())
            
            prevBeam[beamIndex] = beam
            if verbose:
                print("               -> Beam cache contains %i entries" % len(prevBeam.keys()))
                
        # Calculate how much signal needs to be removed...
        toRemove = gain*peakV*beam
        working -= toRemove
        asum = 0.0
        for l in xrange(int(peakXO), int(peakXO)+2):
            if l > peakXO:
                side1 = (peakXO+0.5) - (l-0.5)
            else:
                side1 = (l+0.5) - (peakXO-0.5)
                
            for m in xrange(int(peakYO), int(peakYO)+2):
                if m > peakYO:
                    side2 = (peakYO+0.5) - (m-0.5)
                else:
                    side2 = (m+0.5) - (peakYO-0.5)
                    
                area = side1*side2
                asum += area
                #print 'II', l, m, area, asum
                cleaned[l,m] += gain*area*peakV
                
        if plot:
            pylab.subplot(2, 2, 1)
            pylab.imshow(working+toRemove, origin='lower')
            pylab.title('Before')
            
            pylab.subplot(2, 2, 2)
            pylab.imshow(working, origin='lower')
            pylab.title('After')
            
            pylab.subplot(2, 2, 3)
            pylab.imshow(toRemove, origin='lower')
            pylab.title('CLEAN Comps.')
            
            pylab.subplot(2, 2, 4)
            pylab.imshow(convolve(cleaned, beamClean, mode='same'), origin='lower')
            
            pylab.draw()
            
        if working.max()/working.std() < sigma:
            exitStatus = 'peak is less than %.3f-sigma' % sigma
            
            break
            
    # Summary
    print("Exited after %i iterations with status '%s'" % (i+1, exitStatus))
    
    # Restore
    conv = convolve(cleaned, beamClean, mode='same')
    conv = numpy.ma.array(conv, mask=convMask)
    conv *= ((img-working).max() / conv.max())
    
    if plot:
        # Make an image for comparison purposes if we are verbose
        fig = plt.figure()
        ax1 = fig.add_subplot(2, 2, 1)
        ax2 = fig.add_subplot(2, 2, 2)
        ax3 = fig.add_subplot(2, 2, 3)
        ax4 = fig.add_subplot(2, 2, 4)
        
        c = ax1.imshow(img, extent=(1,-1,-1,1), origin='lower', interpolation='nearest')
        fig.colorbar(c, ax=ax1)
        ax1.set_title('Input')
        
        d = ax2.imshow(conv, extent=(1,-1,-1,1), origin='lower', interpolation='nearest')
        fig.colorbar(d, ax=ax2)
        ax2.set_title('CLEAN Comps.')
        
        e = ax3.imshow(working, extent=(1,-1,-1,1), origin='lower', interpolation='nearest')
        fig.colorbar(e, ax=ax3)
        ax3.set_title('Residuals')
        
        f = ax4.imshow(conv + working, extent=(1,-1,-1,1), origin='lower', interpolation='nearest')
        fig.colorbar(f, ax=ax4)
        ax4.set_title('Final')
        
        plt.show()
        
    if plot:
        pylab.ioff()
        
    # Return
    conv = conv + working
    return conv


def cleanSources(aa, dataDict, aipyImg, srcs, imageInput=None, MapSize=80, MapRes=0.50, MapWRes=0.10, pol='xx', chan=None, gain=0.1, maxIter=150, sigma=2.0, verbose=True, plot=False):
    """
    Given a AIPY antenna array instance, a data dictionary, an AIPY ImgW 
    instance filled with data, and a dictionary of sources, return the CLEAN
    components and the residuals map.  This function uses a CLEAN-like method
    that computes the array beam for each peak in the flux.  Thus the CLEAN 
    loop becomes: 
      1.  Find the peak flux in the residual image
      2.  Compute the systems response to a point source at that location
      3.  Remove the scaled porition of this beam from the residuals
      4.  Go to 1.
    
    This function differs from clean() in that it only cleans localized 
    regions around each source rather than the whole image.  This is
    intended to help the mem() function along.
    
    CLEAN tuning parameters:
      * gain - CLEAN loop gain (default 0.1)
      * maxIter - Maximum number of iterations (default 150)
      * sigma - Threshold in sigma to stop cleaning (default 2.0)
    """
    
    # Sort out the channels to work on
    if chan is None:
        chan = range(dataDict['freq'].size)
        
    # Get a grid of right ascensions and dec values for the image we are working with
    xyz = aipyImg.get_eq(0.0, aa.lat, center=(MapSize,MapSize))
    RA, dec = eq2radec(xyz)
    RA += aa.sidereal_time()
    RA %= (2*numpy.pi)
    top = aipyImg.get_top(center=(MapSize,MapSize))
    az,alt = top2azalt(top)
    
    # Get the list of baselines to generate visibilites for
    baselines = dataDict['bls'][pol]
    
    # Get the actual image out of the ImgW instance
    if imageInput is None:
        img = aipyImg.image(center=(MapSize,MapSize))
    else:
        img = imageInput*1.0
        
    # Setup the arrays to hold the point sources and the residual.
    cleaned = numpy.zeros_like(img)
    working = numpy.zeros_like(img)
    working += img
    
    # Setup the dictionary that will hold the beams as they are computed
    prevBeam = {}
    
    # Estimate the zenith beam response
    psfSrc = {'z': RadioFixedBody(aa.sidereal_time(), aa.lat, jys=1.0, index=0, epoch=aa.date)}
    psfDict = buildSimData(aa, psfSrc, jd=aa.get_jultime(), pols=[pol,], chan=chan, baselines=baselines, flatResponse=True)
    psf = utils.buildGriddedImage(psfDict, MapSize=MapSize, MapRes=MapRes, MapWRes=MapWRes, chan=chan, pol=pol, verbose=verbose)
    psf = psf.image(center=(MapSize,MapSize))
    psf /= psf.max()
    
    # Fit a Guassian to the zenith beam response and use that for the restore beam
    beamCutout = psf[MapSize/2:3*MapSize/2, MapSize/2:3*MapSize/2]
    beamCutout = numpy.where( beamCutout > 0.0, beamCutout, 0.0 )
    h, cx, cy, sx, sy = _fit2DGaussian( beamCutout )
    gauGen = gaussian2d(1.0, MapSize/2+cx, MapSize/2+cy, sx, sy)
    FWHM = int( round( (sx+sy)/2.0 * 2.0*numpy.sqrt(2.0*numpy.log(2.0)) ) )
    beamClean = psf * 0.0
    for i in xrange(beamClean.shape[0]):
        for j in xrange(beamClean.shape[1]):
            beamClean[i,j] = gauGen(i,j)
    beamClean /= beamClean.sum()
    convMask = xyz.mask[0,:,:]
    
    # Go!
    if plot:
        import pylab
        from matplotlib import pyplot as plt
        
        pylab.ion()
        
    for name in srcs.keys():
        src = srcs[name]
        
        # Make sure the source is up
        src.compute(aa)
        print('Source: %s @ %s degrees elevation' % (name, src.alt))
        if src.alt <= 10*numpy.pi/180.0:
            continue
            
        # Locate the approximate position of the source
        srcDist = (src.ra-RA)**2 + (src.dec-dec)**2
        srcPeak = numpy.where( srcDist == srcDist.min() )
        
        # Define the clean box - this is fixed at 2*FWHM in width on each side
        rx0 = srcPeak[0][0] - FWHM/2
        rx1 = rx0 + FWHM + 1
        ry0 = srcPeak[1][0] - FWHM/2
        ry1 = ry0 + FWHM + 1
        
        # Define the background box - this lies outside the clean box and serves
        # as a reference for the background
        X, Y = numpy.indices(working.shape)
        R = numpy.sqrt( (X-srcPeak[0][0])**2 + (Y-srcPeak[1][0])**2 )
        background = numpy.where( (R <= 2*FWHM+3) & (R > 2*FWHM) )
        
        px0 = min(background[0])-1
        px1 = max(background[0])+2
        py0 = min(background[1])-1
        py1 = max(background[1])+2
        
        exitStatus = 'iteration'
        for i in xrange(maxIter):
            # Find the location of the peak in the flux density
            peak = numpy.where( working[rx0:rx1,ry0:ry1] == working[rx0:rx1,ry0:ry1].max() )
            peakX = peak[0][0] + rx0
            peakY = peak[1][0] + ry0
            peakV = working[peakX,peakY]
            
            # Optimize the location
            try:
                peakParams = _fit2DGaussian(working[peakX-FWHM/2:peakX+FWHM/2+1, peakY-FWHM/2:peakY+FWHM/2+1])
            except IndexError:
                peakParams = [peakV, peakX, peakY]
            peakVO = peakParams[0]
            peakXO = peakX - FWHM/2 + peakParams[1]
            peakYO = peakY - FWHM/2 + peakParams[2]
            
            # Quantize to try and keep the computation down without over-simplifiying things
            subpixelationLevel = 5
            peakXO = round(peakXO*subpixelationLevel)/float(subpixelationLevel)
            peakYO = round(peakYO*subpixelationLevel)/float(subpixelationLevel)
            
            # Pixel coordinates to right ascension, dec.
            try:
                peakRA = _interpolateValues(RA, peakXO, peakYO)
            except IndexError:
                peakXO, peakY0 = peakX, peakY
                peakRA = RA[peakX, peakY]
            try:
                peakDec = _interpolateValues(dec, peakXO, peakYO)
            except IndexError:
                peakDec = dec[peakX, peakY]
                
            # Pixel coordinates to az, el
            try:
                peakAz = _interpolateValues(az, peakXO, peakYO)
            except IndexError:
                peakXO, peakYO = peakX, peakY
                peakAz = az[peakX, peakY]
            try:
                peakEl = _interpolateValues(alt, peakX, peakY)
            except IndexError:
                peakEl = alt[peakX, peakY]
                
            if verbose:
                currRA  = deg_to_hms(peakRA * 180/numpy.pi)
                currDec = deg_to_dms(peakDec * 180/numpy.pi)
                currAz  = deg_to_dms(peakAz * 180/numpy.pi)
                currEl  = deg_to_dms(peakEl * 180/numpy.pi)
                
                print("%s - Iteration %i:  Log peak of %.3f at row: %i, column: %i" % (name, i+1, numpy.log10(peakV), peakX, peakY))
                print("               -> RA: %s, Dec: %s" % (currRA, currDec))
                print("               -> az: %s, el: %s" % (currAz, currEl))
                
            # Check for the exit criteria
            if peakV < 0:
                exitStatus = 'peak value is negative'
                
                break
            
            # Find the beam index and see if we need to compute the beam or not
            beamIndex = (int(peakXO*subpixelationLevel), int(peakYO*subpixelationLevel))
            try:
                beam = prevBeam[beamIndex]
                
            except KeyError:
                if verbose:
                    print("               -> Computing beam(s)")
                    
                beamSrc = {'Beam': RadioFixedBody(peakRA, peakDec, jys=1.0, index=0, epoch=aa.date)}
                beamDict = buildSimData(aa, beamSrc, jd=aa.get_jultime(), pols=[pol,], chan=chan, baselines=baselines, flatResponse=True)
                beam = utils.buildGriddedImage(beamDict, MapSize=MapSize, MapRes=MapRes, MapWRes=MapWRes, chan=chan, pol=pol, verbose=verbose)
                beam = beam.image(center=(MapSize,MapSize))
                beam /= beam.max()
                print("                  ", beam.mean(), beam.min(), beam.max(), beam.sum())
                
                prevBeam[beamIndex] = beam
                if verbose:
                    print("               -> Beam cache contains %i entries" % len(prevBeam.keys()))
                    
            # Calculate how much signal needs to be removed...
            toRemove = gain*peakV*beam
            working -= toRemove
            asum = 0.0
            for l in xrange(int(peakXO), int(peakXO)+2):
                if l > peakXO:
                    side1 = (peakXO+0.5) - (l-0.5)
                else:
                    side1 = (l+0.5) - (peakXO-0.5)
                    
                for m in xrange(int(peakYO), int(peakYO)+2):
                    if m > peakYO:
                        side2 = (peakYO+0.5) - (m-0.5)
                    else:
                        side2 = (m+0.5) - (peakYO-0.5)
                        
                    area = side1*side2
                    asum += area
                    #print 'II', l, m, area, asum
                    cleaned[l,m] += gain*area*peakV
                    
            if plot:
                try:
                    pylab.subplot(2, 2, 1)
                    pylab.imshow((working+toRemove)[px0:px1,py0:py1], origin='lower', interpolation='nearest')
                    pylab.title('Before')
                    
                    pylab.subplot(2, 2, 2)
                    pylab.imshow(working[px0:px1,py0:py1], origin='lower', interpolation='nearest')
                    pylab.title('After')
                    
                    pylab.subplot(2, 2, 3)
                    pylab.imshow(toRemove[px0:px1,py0:py1], origin='lower', interpolation='nearest')
                    pylab.title('Removed')
                    
                    pylab.subplot(2, 2, 4)
                    pylab.imshow(convolve(cleaned, beamClean, mode='same')[px0:px1,py0:py1], origin='lower', interpolation='nearest')
                    pylab.title('CLEAN Comps.')
                except:
                    pass
                    
                try:
                    st.set_text('%s @ %i' % (name, i+1))
                except NameError:
                    st = pylab.suptitle('%s @ %i' % (name, i+1))
                pylab.draw()
                
            if numpy.abs(numpy.max(working[rx0:rx1,ry0:ry1])-numpy.median(working[background]))/rStd(working[background]) <= sigma:
                exitStatus = 'peak is less than %.3f-sigma' % sigma
                
                break
                
        # Summary
        print("Exited after %i iterations with status '%s'" % (i+1, exitStatus))
        
    # Restore
    conv = convolve(cleaned, beamClean, mode='same')
    conv = numpy.ma.array(conv, mask=convMask)
    conv *= ((img-working).max() / conv.max())
    
    if plot:
        # Make an image for comparison purposes if we are verbose
        fig = plt.figure()
        ax1 = fig.add_subplot(2, 2, 1)
        ax2 = fig.add_subplot(2, 2, 2)
        ax3 = fig.add_subplot(2, 2, 3)
        ax4 = fig.add_subplot(2, 2, 4)
        
        c = ax1.imshow(img, extent=(1,-1,-1,1), origin='lower', interpolation='nearest')
        fig.colorbar(c, ax=ax1)
        ax1.set_title('Input')
        
        d = ax2.imshow(conv, extent=(1,-1,-1,1), origin='lower', interpolation='nearest')
        fig.colorbar(d, ax=ax2)
        ax2.set_title('CLEAN Comps.')
        
        e = ax3.imshow(working, extent=(1,-1,-1,1), origin='lower', interpolation='nearest')
        fig.colorbar(e, ax=ax3)
        ax3.set_title('Residuals')
        
        f = ax4.imshow(conv + working, extent=(1,-1,-1,1), origin='lower', interpolation='nearest')
        fig.colorbar(f, ax=ax4)
        ax4.set_title('Final')
        
        plt.show()
        
    if plot:
        pylab.ioff()
        
    # Return
    return conv, working


def _minorCycle(img, beam, gain=0.2, nIter=150):
    """
    Function for performing the minor cycle part of lsq().
    """
    
    cleaned = img*0.0
    working = img*1.0
    
    for i in xrange(nIter):
        # Find the location of the peak in the flux density
        aw = numpy.abs( working )
        peak = numpy.where( aw == aw.max() )
        peakX = peak[0][0]
        peakY = peak[1][0]
        peakV = working[peakX,peakY]
        
        if numpy.abs(peakV - working.mean()) < 2*working.std():
            break
            
        # Build the beam
        beam2 = numpy.roll(beam,  peakX-beam.shape[0]/2, axis=0)
        beam2 = numpy.roll(beam2, peakY-beam.shape[1]/2, axis=1) 
        
        # Calculate how much signal needs to be removed...
        toRemove = gain*peakV*beam2
        working -= toRemove
        cleaned[peakX,peakY] += toRemove.sum()
        
    return  cleaned + working


def lsq(aa, dataDict, aipyImg, imageInput=None, MapSize=80, MapRes=0.50, MapWRes=0.10, pol='xx', chan=None, gain=0.05, maxIter=150, rtol=1e-9, verbose=True, plot=False):
    """
    Given a AIPY antenna array instance, a data dictionary, and an AIPY ImgW 
    instance filled with data, return a deconvolved image.  This function 
    implements a least squares deconvolution.
    
    Least squares tuning parameters:
      * gain - least squares loop gain (default 0.05)
      * maxIter - Maximum number of iteration (default 150)
      * rtol - Minimum change in the residual RMS between iterations
               (default 1e-9)
    """
    
    # Sort out the channels to work on
    if chan is None:
        chan = range(dataDict['freq'].size)
        
    # Get a grid of right ascensions and dec values for the image we are working with
    xyz = aipyImg.get_eq(aa.sidereal_time(), aa.lat, center=(MapSize,MapSize))
    top = aipyImg.get_top(center=(MapSize,MapSize))
    ra, dec = eq2radec(xyz)
    
    # Get the list of baselines to generate visibilites for
    baselines = dataDict['bls'][pol]
    
    # Estimate the zenith beam response
    psfSrc = {'z': RadioFixedBody(aa.sidereal_time(), aa.lat, jys=1.0, index=0, epoch=aa.date)}
    psfDict = buildSimData(aa, psfSrc, jd=aa.get_jultime(), pols=[pol,], chan=chan, baselines=baselines, flatResponse=True)
    psf = utils.buildGriddedImage(psfDict, MapSize=MapSize, MapRes=MapRes, MapWRes=MapWRes, chan=chan, pol=pol, verbose=verbose)
    psf = psf.image(center=(MapSize,MapSize))
    psf /= psf.max()
    
    # Fit a Guassian to the zenith beam response and use that for the restore beam
    beamCutout = psf[MapSize/2:3*MapSize/2, MapSize/2:3*MapSize/2]
    beamCutout = numpy.where( beamCutout > 0.0, beamCutout, 0.0 )
    h, cx, cy, sx, sy = _fit2DGaussian( beamCutout )
    gauGen = gaussian2d(1.0, MapSize/2+cx, MapSize/2+cy, sx, sy)
    FWHM = int( round( (sx+sy)/2.0 * 2.0*numpy.sqrt(2.0*numpy.log(2.0)) ) )
    beamClean = psf * 0.0
    for i in xrange(beamClean.shape[0]):
        for j in xrange(beamClean.shape[1]):
            beamClean[i,j] = gauGen(i,j)
    beamClean /= beamClean.sum()
    convMask = xyz.mask[0,:,:]
    
    # Get the actual image out of the ImgW instance
    if imageInput is None:
        img = aipyImg.image(center=(MapSize,MapSize))
    else:
        img = imageInput*1.0
        
    # Build the initial model
    mdl = img*40
    mdl[numpy.where(mdl < 0)] = 0
    mdl[numpy.where(ra.mask == 1)] = 0
    
    # Determine the overall image->model scale factor
    bSrcs = {}
    rChan = [chan[0], chan[-1]]
    bSrcs['zenith'] = RadioFixedBody(aa.sidereal_time(), aa.lat, name='zenith', jys=1, index=0)
    simDict = buildSimData(aa, bSrcs, jd=aa.get_jultime(), pols=[pol,], chan=rChan, baselines=baselines, flatResponse=True)
    simImg = utils.buildGriddedImage(simDict, MapSize=MapSize, MapRes=MapRes, MapWRes=MapWRes, chan=rChan, pol=pol, verbose=verbose)
    simImg = simImg.image(center=(MapSize,MapSize))
    
    simToModel = 1.0 / simImg.max()
    modelToSim = simImg.max() / 1.0
    
    # Go!
    if plot:
        import pylab
        from matplotlib import pyplot as plt
        
        pylab.ion()
        
    rChan = [chan[0], chan[-1]]
    diff = img - mdl
    diffScaled = 0.0*diff/gain
    oldModel = mdl
    oldRMS = diff.std()*1e6
    oldDiff = diff*0.0
    rHist = []
    exitStatus = 'iteration'
    for k in xrange(maxIter):
        ## Update the model image but don't allow negative flux
        mdl += diffScaled * gain
        mdl[numpy.where( mdl <= 0 )] = 0.0
        
        ## Convert the model image to an ensemble of point sources for forward 
        ## modeling
        bSrcs = {}
        for i in xrange(mdl.shape[0]):
            for j in xrange(mdl.shape[1]):
                if dec.mask[i,j]:
                    continue
                if mdl[i,j] <= 0:
                    continue
                    
                nm = '%i-%i' % (i,j)
                bSrcs[nm] = RadioFixedBody(ra[i,j], dec[i,j], name=nm, jys=mdl[i,j], index=0, epoch=aa.date)
                
        ## Model the visibilities
        simDict = buildSimData(aa, bSrcs, jd=aa.get_jultime(), pols=[pol,], chan=rChan, baselines=baselines, flatResponse=True)
        
        ## Form the simulated image
        simImg = utils.buildGriddedImage(simDict, MapSize=MapSize, MapRes=MapRes, MapWRes=MapWRes, chan=rChan, pol=pol, verbose=verbose)
        simImg = simImg.image(center=(MapSize,MapSize))
        
        ## Difference the image and the simulated image and scale it to the 
        ## model's peak flux
        diff = img - simImg
        diff2 = _minorCycle(diff, beamClean, gain=0.1, nIter=2000)
        
        ## Compute the RMS and create an appropriately scaled version of the model
        RMS = diff.std()
        mdl2 = mdl*modelToSim
        
        ## Status report
        if verbose:
            print("Iteration %i:  %i sources used, RMS is %.4e" % (k+1, len(bSrcs.keys()), RMS))
            print("               -> maximum residual: %.4e (%.3f%% of peak)" % (diff.max(), 100.0*diff.max()/img.max()))
            print("               -> minimum residual: %.4e (%.3f%% of peak)" % (diff.min(), 100.0*diff.min()/img.max()))
            print("               -> delta RMS: %.4e (%.3f%%)" % (RMS-oldRMS, 100.0*(RMS-oldRMS)/RMS))
            
        ## Make the cleaned residuals map ready for updating the model
        diff = diff2
        diffScaled = diff * simToModel
        
        ## Has the RMS gone up?  If so, it is time to exit.  But first, restore 
        ## the previous iteration
        if RMS-oldRMS > 0:
            mdl = oldModel
            diff = oldDiff
            
            exitStatus = 'residuals'
            
            break
            
        ## Is the RMS still changing in an acceptable manner?
        if abs(RMS-oldRMS) < rtol:
            # No need to go back a step			
            #mdl = oldModel
            #diff = oldDiff
            
            exitStatus = 'tolerance'
            
            break
            
        ## Save the current iteration as the previous state
        rHist.append(RMS)
        oldRMS = RMS
        oldModel = mdl
        oldDiff = diff
        
        if plot:
            pylab.subplot(3, 2, 1)
            pylab.imshow(img, origin='lower', interpolation='nearest', vmin=img.min(), vmax=img.max())
            pylab.subplot(3, 2, 2)
            pylab.imshow(simImg, origin='lower', interpolation='nearest', vmin=img.min(), vmax=img.max())
            pylab.subplot(3, 2, 3)
            pylab.imshow(diff, origin='lower', interpolation='nearest')
            pylab.subplot(3, 2, 4)
            pylab.imshow(mdl, origin='lower', interpolation='nearest')
            pylab.subplot(3, 1, 3)
            pylab.cla()
            pylab.semilogy(rHist)
            pylab.draw()
            
    # Summary
    print("Exited after %i iterations with status '%s'" % (k+1, exitStatus))
    
    # Restore
    conv = convolve(mdl2, beamClean, mode='same')
    conv = numpy.ma.array(conv, mask=convMask)
    
    if plot:
        # Make an image for comparison purposes if we are verbose
        fig = plt.figure()
        ax1 = fig.add_subplot(2, 2, 1)
        ax2 = fig.add_subplot(2, 2, 2)
        ax3 = fig.add_subplot(2, 2, 3)
        ax4 = fig.add_subplot(2, 2, 4)
        
        c = ax1.imshow(img, extent=(1,-1,-1,1), origin='lower', interpolation='nearest')
        fig.colorbar(c, ax=ax1)
        ax1.set_title('Input')
        
        d = ax2.imshow(simImg, extent=(1,-1,-1,1), origin='lower', interpolation='nearest')
        fig.colorbar(d, ax=ax2)
        ax2.set_title('Realized Model')
        
        e = ax3.imshow(diff, extent=(1,-1,-1,1), origin='lower', interpolation='nearest')
        fig.colorbar(e, ax=ax3)
        ax3.set_title('Residuals')
        
        f = ax4.imshow(conv + diff, extent=(1,-1,-1,1), origin='lower', interpolation='nearest')
        fig.colorbar(f, ax=ax4)
        ax4.set_title('Final')
        
        plt.show()
        
    if plot:
        pylab.ioff()
        
    return conv + diff



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import argparse

from lsl.misc import wisdom


def main(args):
    do_fftw = True if not args.pyfftw_only else False
    do_pyfftw = True if not args.fftw_only else False
    
    if args.show:
        wisdom.show(FFTW=do_fftw, PyFFTW=do_pyfftw)
    else:
        wisdom.make(FFTW=do_fftw, PyFFTW=do_pyfftw)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='build LSL-specific FFTW and PyFFTW wisdom files', 
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
        )
    parser.add_argument('-f', '--fftw-only', action='store_true', 
                        help='build/show only FFTW wisdom only')
    parser.add_argument('-p', '--pyfftw-only', action='store_true', 
                        help='build/show only PyFFTW wisdom only')
    parser.add_argument('-s', '--show', action='store_true', 
                        help='show information about the avaliable LSL wisdom')
    args = parser.parse_args()
    main(args)
    



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Read and plot the NOSTA_MAPPER table in a FITS IDI file writen by 
lsl.writer.fitsidi if it exists."""

import sys
import numpy
import pyfits

from lsl.common import stations

import matplotlib.pyplot as plt
from matplotlib.ticker import NullFormatter


def main(args):
    # Set the station
    station = stations.lwa1
    antennas = station.getAntennas()[0::2]
    stands = []
    for ant in antennas:
        stands.append(ant.stand.id)

    # Open the FITS file for reading
    hdu = pyfits.open(args[0])

    # This try...except block catches files that don't use the stand mapper 
    # feature in lsl.writer.fitsidi.  If a NOSTA_MAPPER table doesn't exist,
    # the script fall back on the ARRAY_GEOMETRY table to gather the stand 
    # numbers.
    try:
        mapper = hdu['NOSTA_MAPPER']
        
        nosta = mapper.data.field('NOSTA')
        noact = mapper.data.field('NOACT')
        anname = mapper.data.field('ANNAME')
    except KeyError:
        ag = hdu['ARRAY_GEOMETRY']

        nosta = ag.data.field('NOSTA')
        noact = ag.data.field('NOSTA')
        anname = ag.data.field('ANNAME')

    # Print the stand mapping out
    print "Stand Mapping:"
    for sta,act,name in zip(nosta, noact, anname):
        print "  %3i -> %s (stand %3i)" % (sta, name, act)
    
    # Load in the positions of all of the stands
    xyz = numpy.zeros((len(antennas), 3))
    i = 0
    for ant in antennas:
        xyz[i,0] = ant.stand.x
        xyz[i,1] = ant.stand.y
        xyz[i,2] = ant.stand.z
        i += 1

    # Begin the plot
    fig = plt.figure()
    ax1 = fig.add_subplot(1, 1, 1)
    ax2 = plt.axes([0.75, 0.75, 0.1, 0.1])
    
    ## Part 1 - Station between -80 and 80 (the section inside the fence)
    c = ax1.scatter(xyz[:,0], xyz[:,1], c=xyz[:,2], s=40.0, alpha=0.50)
    for mSta,sAct in zip(nosta, noact):
        i = stands.index(sAct)
        ax1.plot(xyz[i,0], xyz[i,1], marker='x', linestyle=' ', alpha=1.0, color='k')
        ax1.text(xyz[i,0], xyz[i,1], ' %i' % mSta)
    ax1.set_xlabel('$\Delta$X [E-W; m]')
    ax1.set_xlim([-80, 80])
    ax1.set_ylabel('$\Delta$Y [N-S; m]')
    ax1.set_ylim([-80, 80])
    
    ## Part 2 - The outlier as inset axes
    ax2.scatter(xyz[:,0], xyz[:,1], c=xyz[:,2], s=40.0, alpha=0.50)
    for mSta,sAct in zip(nosta, noact):
        i = stands.index(sAct)
        ax2.plot(xyz[i,0], xyz[i,1], marker='x', linestyle=' ', alpha=1.0, color='k')
        ax2.text(xyz[i,0], xyz[i,1], ' %i' % mSta)
    ax2.set_title('RTA (Outlier)')
    ax2.set_xlim([335, 345])
    ax2.set_ylim([10, 20])
    ax2.xaxis.set_major_formatter( NullFormatter() )
    ax2.yaxis.set_major_formatter( NullFormatter() )
    
    ### Part 3 - Outlier direction indicator
    #x0 = 62.0
    #y0 = 62.0
    #i = stands.index(258)
    #dx = xyz[i,0] - x0
    #dy = xyz[i,1] - y0
    #ax1.arrow(x0, y0, dx*8/dx, dy*8/dx, linewidth=2.0)
    #ax1.text(x0+5, y0, '(%i m)' % numpy.sqrt(dx**2 + dy**2))

    plt.show()

if __name__ == "__main__":
    main(sys.argv[1:])



# -*- coding: utf-8 -*-

# Python3 compatiability
from __future__ import absolute_import

"""Unit test suite for the lsl package."""

import unittest

from . import test_paths
from . import test_astro
from . import test_skymap
from . import test_transform
from . import test_mathutil
from . import test_nec_util
from . import test_catalog
from . import test_dp
from . import test_stations
from . import test_robust
from . import test_stattests
from . import test_reader
from . import test_reader_adp
from . import test_buffer
from . import test_ldp
from . import test_ldp_adp
from . import test_uvUtils
from . import test_fx
from . import test_filterbank
from . import test_fakedata
from . import test_simdp
from . import test_simvis
from . import test_geodesy
from . import test_fitsidi
from . import test_uvfits
from . import test_tsfits
from . import test_sdfits
from . import test_vdif
from . import test_measurementset
from . import test_beamformer
from . import test_imaging
from . import test_progress
from . import test_busy
from . import test_mcs
from . import test_mcs_adp
from . import test_sdf
from . import test_sdf_adp
from . import test_meta
from . import test_meta_adp
from . import test_parser
from . import test_idf


__revision__  = "$Rev: 2751 $"
__version__   = "0.5"
__author__    = "D.L.Wood"
__maintainer__ = "Jayce Dowell"


class lsl_tests(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the package unit tests."""

    def __init__(self):
        """Setup the lsl package unit test suite."""

        unittest.TestSuite.__init__(self)
        
        self.addTest(test_paths.paths_test_suite())
        self.addTest(test_astro.astro_test_suite())
        self.addTest(test_skymap.skymap_test_suite())
        self.addTest(test_transform.transform_test_suite())
        self.addTest(test_mathutil.mathutil_test_suite())
        self.addTest(test_nec_util.nec_util_test_suite())
        self.addTest(test_catalog.catalog_test_suite())
        self.addTest(test_dp.dp_test_suite())
        self.addTest(test_stations.stations_test_suite())
        self.addTest(test_robust.robust_test_suite())
        self.addTest(test_stattests.stattests_test_suite())
        self.addTest(test_reader.reader_test_suite())
        self.addTest(test_reader_adp.reader_adp_test_suite())
        self.addTest(test_buffer.buffer_test_suite())
        self.addTest(test_ldp.ldp_test_suite())
        self.addTest(test_ldp_adp.ldp_adp_test_suite())
        self.addTest(test_uvUtils.uvUtils_test_suite())
        self.addTest(test_fx.fx_test_suite())
        self.addTest(test_filterbank.filterbank_test_suite())
        self.addTest(test_fakedata.fakedata_test_suite())
        self.addTest(test_simdp.simdp_test_suite())
        self.addTest(test_simvis.simvis_test_suite())
        self.addTest(test_geodesy.geodesy_test_suite())
        self.addTest(test_fitsidi.fitsidi_test_suite())
        self.addTest(test_uvfits.uvfits_test_suite())
        self.addTest(test_tsfits.tsfits_test_suite())
        self.addTest(test_sdfits.sdfits_test_suite())
        self.addTest(test_vdif.vdif_test_suite())
        self.addTest(test_measurementset.measurementset_test_suite())
        self.addTest(test_beamformer.beamformer_test_suite())
        self.addTest(test_imaging.imaging_test_suite())
        self.addTest(test_progress.progress_test_suite())
        self.addTest(test_busy.busy_test_suite())
        self.addTest(test_mcs.mcs_test_suite())
        self.addTest(test_mcs_adp.mcs_adp_test_suite())
        self.addTest(test_sdf.sdf_test_suite())
        self.addTest(test_sdf_adp.sdf_adp_test_suite())
        self.addTest(test_meta.metabundle_test_suite())
        self.addTest(test_meta_adp.metabundle_adp_test_suite())
        self.addTest(test_parser.parser_test_suite())
        self.addTest(test_idf.idf_test_suite())


def main(opts=None, args=None):
    """Function to call all of the lsl tests."""

    if opts is not None:
        if opts.verbose:
            level = 2
        else:
            level = 1
    else:
        level = 2
            
    suite = lsl_tests()
    runner = unittest.TextTestRunner(verbosity = level)
    runner.run(suite)


if __name__  == '__main__':
    import optparse
    
    parser = optparse.OptionParser(usage = "python %prog [options]", description = __doc__)
    parser.add_option("-v", "--verbose", action = "store_true", dest = "verbose", default = False,
        help = "extra print output")
    (opts, args) = parser.parse_args()

    main(opts, args)



# -*- coding: utf-8 -*-

# Python3 compatiability
from __future__ import print_function
import sys
if sys.version_info > (3,):
    xrange = range
    
"""
Simple self-calibration module for correlated TBW and TBN data.  The 
supported self-calibration methods are:
  * phase-only
  * amplitude and phase
  * delay-only
  * amplitude and delay
  * delay/phase offset
  * amplitude and delay/phase offset

..versionchanged:: 0.6.3
    Reworked the module to make it more versatile
    
..versionadded:: 0.5.5
"""

import numpy

from lsl.statistics import robust

__version__ = '0.2'
__revision__ = '$Rev: 2457 $'
__all__ = ['phaseOnly', 'delayOnly', 'delayAndPhase', '__version__', '__revision__', '__all__']


def _scaleData(dataDict, amps, delays, phaseOffsets):
    """
    Apply a set of antenna-based real gain value, phase delays in ns, phase 
    offsets in radians to a data dictionary.  Returned the new scaled and 
    delayed dictionary.
    """

    import copy

    # Build the data dictionary to hold the scaled and delayed data
    sclUVData = {'freq': (dataDict['freq']).copy(), 'uvw': {}, 'vis': {}, 'wgt': {}, 'msk': {}, 'bls': {}, 'jd': {}}
    if dataDict['isMasked']:
        sclUVData['isMasked'] = True
    else:
        sclUVData['isMasked'] = False
    fq = dataDict['freq'] / 1e9
    
    cGains = []
    for i in xrange(len(amps)):
        cGains.append( amps[i]*numpy.exp(2j*numpy.pi*fq*delays[i] + 1j*phaseOffsets[i]) )

    # Apply the scales and delays for all polarization pairs found in the original data
    for pol in dataDict['vis'].keys():
        sclUVData['bls'][pol] = []
        sclUVData['uvw'][pol] = []
        sclUVData['vis'][pol] = []
        sclUVData['wgt'][pol] = copy.copy(dataDict['wgt'][pol])
        sclUVData['msk'][pol] = copy.copy(dataDict['msk'][pol])
        sclUVData['jd'][pol] = copy.copy(dataDict['jd'][pol])

        for (i,j),uvw,vis in zip(dataDict['bls'][pol], dataDict['uvw'][pol], dataDict['vis'][pol]):
            sclUVData['bls'][pol].append( (i,j) )
            sclUVData['uvw'][pol].append( uvw )
            sclUVData['vis'][pol].append( vis*cGains[j].conj()*cGains[i] )

    return sclUVData


def _buildAmplitudeA(aa, dataDict, simDict, chan, pol, refAnt=0):
    """
    Build the matrix A for amplitude correction.
    """
    
    # Get the baseline and stand counts
    nBLs = len(dataDict['bls'][pol])
    nStands = len(aa.ants)
    
    # Frequency in GHz so that the delays can be in ns
    fq = dataDict['freq'][chan] / 1e9
    
    A = numpy.zeros((nBLs, nStands))
    for i,(l,m) in enumerate(dataDict['bls'][pol]):
        A[i,l] = 1.0
        A[i,m] = 1.0
        
    A = numpy.matrix(A)
    return A


def _buildAmplitudeC(aa, dataDict, simDict, chan, pol, refAnt=0):
    """
    Build the matrix C for the amplitude correction.
    """
    
    # Get the baseline and stand counts
    nBLs = len(dataDict['bls'][pol])
    nStands = len(aa.ants)
    
    # Frequency in GHz so that the delays can be in ns
    fq = dataDict['freq'][chan] / 1e9
    
    obsVis = []
    for vis in dataDict['vis'][pol]:
        obsVis.append( numpy.array(vis[chan]) )
    obsVis = numpy.array(obsVis)
    
    simVis = []
    for vis in simDict['vis'][pol]:
        simVis.append( numpy.array(vis[chan]) )
    simVis = numpy.array(simVis)
    
    C = numpy.log(numpy.abs(simVis / obsVis))
    
    Cp = numpy.zeros(nBLs)
    for i in xrange(C.shape[0]):
        Cp[i] = robust.mean(C[i,:])
    
    return Cp


def _buildPhaseOnlyA(aa, dataDict, simDict, chan, pol, refAnt=0):
    """
    Build the matrix A for phase correction with a frequency independent
    phase.
    """
    
    # Get the baseline and stand counts
    nBLs = len(dataDict['bls'][pol])
    nStands = len(aa.ants)
    
    # Frequency in GHz so that the delays can be in ns
    fq = dataDict['freq'][chan] / 1e9
    
    A = numpy.zeros((nBLs*fq.size, nStands-1))
    for i in xrange(fq.size):
        for j,(l,m) in enumerate(dataDict['bls'][pol]):
            if l < refAnt:
                A[j+i*nBLs,l]   =  1.0
            elif l > refAnt:
                A[j+i*nBLs,l-1] =  1.0
            else:
                pass
                
            if m < refAnt:
                A[j+i*nBLs,m]   = -1.0
            elif m > refAnt:
                A[j+i*nBLs,m-1] = -1.0
            else:
                pass
        
    A = numpy.matrix(A)
    return A


def _buildPhaseOnlyC(aa, dataDict, simDict, chan, pol, refAnt=0):
    """
    Build the matrix C for phase correction with a frequency independent
    phase.
    """
    
    # Get the baseline and stand counts
    nBLs = len(dataDict['bls'][pol])
    nStands = len(aa.ants)
    
    # Frequency in GHz so that the delays can be in ns
    fq = dataDict['freq'][chan] / 1e9
    
    obsVis = []
    for vis in dataDict['vis'][pol]:
        obsVis.append( numpy.array(vis[chan]) )
    obsVis = numpy.array(obsVis)
    
    simVis = []
    for vis in simDict['vis'][pol]:
        simVis.append( numpy.array(vis[chan]) )
    simVis = numpy.array(simVis)
    
    C = numpy.angle(simVis / obsVis)
    
    Cp = numpy.zeros(nBLs*fq.size)
    for i in xrange(fq.size):
        for j in xrange(C.shape[0]):
            Cp[j+i*nBLs] = C[j,i]
    
    return Cp


def _buildDelayOnlyA(aa, dataDict, simDict, chan, pol, refAnt=0):
    """
    Build the matrix A for phase correction with a delay.
    """
    
    # Get the baseline and stand counts
    nBLs = len(dataDict['bls'][pol])
    nStands = len(aa.ants)
    
    # Frequency in GHz so that the delays can be in ns
    fq = dataDict['freq'][chan] / 1e9
    
    A = numpy.zeros((nBLs*fq.size, nStands-1))
    for i in xrange(fq.size):
        for j,(l,m) in enumerate(dataDict['bls'][pol]):
            if l < refAnt:
                A[j+i*nBLs,l]   =  2*numpy.pi*fq[i]
            elif l > refAnt:
                A[j+i*nBLs,l-1] =  2*numpy.pi*fq[i]
            else:
                pass
                
            if m < refAnt:
                A[j+i*nBLs,m]   = -2*numpy.pi*fq[i]
            elif m > refAnt:
                A[j+i*nBLs,m-1] = -2*numpy.pi*fq[i]
            else:
                pass
        
    A = numpy.matrix(A)
    return A


def _buildDelayOnlyC(aa, dataDict, simDict, chan, pol, refAnt=0):
    """
    Build the matrix C for phase correction with a delay.
    """
    
    # Get the baseline and stand counts
    nBLs = len(dataDict['bls'][pol])
    nStands = len(aa.ants)
    
    # Frequency in GHz so that the delays can be in ns
    fq = dataDict['freq'][chan] / 1e9
    
    obsVis = []
    for vis in dataDict['vis'][pol]:
        obsVis.append( numpy.array(vis[chan]) )
    obsVis = numpy.array(obsVis)
    
    simVis = []
    for vis in simDict['vis'][pol]:
        simVis.append( numpy.array(vis[chan]) )
    simVis = numpy.array(simVis)
    
    C = numpy.angle(simVis / obsVis)
    
    Cp = numpy.zeros(nBLs*fq.size)
    for i in xrange(fq.size):
        for j in xrange(C.shape[0]):
            Cp[j+i*nBLs] = C[j,i]
    
    return Cp


def _buildDelayAndPhaseA(aa, dataDict, simDict, chan, pol, refAnt=0):
    """
    Build the matrix A for phase correction with a delay and a phase offset.
    """
    
    # Get the baseline and stand counts
    nBLs = len(dataDict['bls'][pol])
    nStands = len(aa.ants)
    
    # Frequency in GHz so that the delays can be in ns
    fq = dataDict['freq'][chan] / 1e9
    
    A = numpy.zeros((nBLs*fq.size, 2*(nStands-1)))
    for i in xrange(fq.size):
        for j,(l,m) in enumerate(dataDict['bls'][pol]):
            if l < refAnt:
                A[j+i*nBLs,l]   =  2*numpy.pi*fq[i]
                A[j+i*nBLs,l+(nStands-1)] = 1.0
            elif l > refAnt:
                A[j+i*nBLs,l-1] =  2*numpy.pi*fq[i]
                A[j+i*nBLs,l-1+(nStands-1)] = 1.0
                
            else:
                pass
                
            if m < refAnt:
                A[j+i*nBLs,m]   = -2*numpy.pi*fq[i]
                A[j+i*nBLs,m+(nStands-1)] = -1.0
            elif m > refAnt:
                A[j+i*nBLs,m-1] = -2*numpy.pi*fq[i]
                A[j+i*nBLs,m-1+(nStands-1)] = -1.0
            else:
                pass
        
    A = numpy.matrix(A)
    return A


def _buildDelayAndPhaseC(aa, dataDict, simDict, chan, pol, refAnt=0):
    """
    Build the matrix C for phase correction with a delay and a phase offset.
    """
    
    # Get the baseline and stand counts
    nBLs = len(dataDict['bls'][pol])
    nStands = len(aa.ants)
    
    # Frequency in GHz so that the delays can be in ns
    fq = dataDict['freq'][chan] / 1e9
    
    obsVis = []
    for vis in dataDict['vis'][pol]:
        obsVis.append( numpy.array(vis[chan]) )
    obsVis = numpy.array(obsVis)
    
    simVis = []
    for vis in simDict['vis'][pol]:
        simVis.append( numpy.array(vis[chan]) )
    simVis = numpy.array(simVis)
    
    C = numpy.angle(simVis / obsVis)
    
    Cp = numpy.zeros(nBLs*fq.size)
    for i in xrange(fq.size):
        for j in xrange(C.shape[0]):
            Cp[j+i*nBLs] = C[j,i]
    
    return Cp


def _selfCal(aa, dataDict, simDict, chan, pol, refAnt=0, nIter=30, amplitude=False, phaseOnly=False, delayOnly=False, delayAndPhase=False, amplitudeCutoff=1.001, phaseCutoff=0.01, delayCutoff=0.2, verbose=True):
    """
    Function used to perform a variety of self-calibration strategies on 
    data stored in a readUVData dictionary and a model sky stored in a 
    lsl.sim.vis.buildSimSky dictionary for a given polarization and 
    channel(s).
    
    The supported self-cal. schemes are:
      * phase-only
      * amplitude and phase
      * delay-only
      * amplitude and delay
      * delay/phase offset
      * amplitude and delay/phase offset
    
    The function exits when either the maximum number of iterations is 
    reached (nIter) or the maximum absolute value of the relevant quantity
    drops below the quantity's cutoff value.  The cutoff keywords and 
    their default values are:
      * amplitudeCutoff - 1.001, 
      * phaseCutoff - 0.01 radians, 
      * delayCutoff - 0.2 ns (0.01 radians over 10 MHz),
    """

    # Make sure we have the right polarization
    if pol not in dataDict['bls'].keys() and pol.lower() not in dataDict['bls'].keys():
        raise RuntimeError("Data dictionary does not have data for polarization '%s'" % pol)
    if pol not in simDict['bls'].keys() and pol.lower() not in simDict['bls'].keys():
        raise RuntimeError("Simulation dictionary does not have data for polarization '%s'" % pol)

    # Make sure that `chan' is an array by trying to find its length
    try:
        junk = len(chan)
    except TypeError:
        chan = [chan]
        
    N = len(aa.ants)
    found = False
    tauNames = []
    phsNames = []
    if refAnt != 0:
        origRefAnt = refAnt
        for i,ant in enumerate(aa.ants):
            if origRefAnt == ant.stand:
                refAnt = i
                found = True
            else:
                tauNames.append('tau%i' % ant.stand)
                phsNames.append('phs%i' % ant.stand)
    else:
        found = True
    if not found:
        raise RuntimeError("Stand #%i not found in the array provided" % refAnt)
    if verbose:
        print("Using antenna #%i as a reference (Stand #%i)" % (refAnt, aa.ants[refAnt].stand))
        
    # Frequency in GHz so that the delays can be in ns
    fq = dataDict['freq'][chan] / 1e9
    
    tempGains = numpy.ones(N)
    tempDelays = numpy.zeros(N)
    tempPhaseOffsets = numpy.zeros(N)
    for i in xrange(nIter):
        #
        # Amplitude
        #
        if amplitude:
            if verbose:
                print('  %iA' % (i+1,))
                
            A = _buildAmplitudeA(aa, dataDict, simDict, chan, pol, refAnt=refAnt)
            C = _buildAmplitudeC(aa, dataDict, simDict, chan, pol, refAnt=refAnt)
            
            good = numpy.where( numpy.isfinite(C) == 1 )[0]
            A = A[good,:]
            C = C[good]
            
            bestGains, resid, rank, s = numpy.linalg.lstsq(A, C)
            resid = numpy.array(C - numpy.dot(A, bestGains)).ravel()
            resid = (C**2).sum(), (resid**2).sum()
            bestGains = numpy.exp(bestGains)
            tempGains *= bestGains
            
            valid = numpy.where( numpy.abs(bestGains) < 1e6 )[0]
            metric = (numpy.abs(bestGains[valid])).max()
            if verbose:
                print('    ', metric)
            if metric < amplitudeCutoff:
                amplitude = False
                
            dataDict = _scaleData(dataDict, bestGains, numpy.zeros_like(bestGains), numpy.zeros_like(bestGains))
        
        #
        # Delay and/or phase
        #
        if phaseOnly:
            if verbose:
                print('  %iP' % (i+1,))
                
            A = _buildPhaseOnlyA(aa, dataDict, simDict, chan, pol, refAnt=refAnt)
            C = _buildPhaseOnlyC(aa, dataDict, simDict, chan, pol, refAnt=refAnt)
            
            good = numpy.where( numpy.isfinite(C) == 1 )[0]
            A = A[good,:]
            C = C[good]
            
            bestPhaseOffsets, resid, rank, s = numpy.linalg.lstsq(A, C)
            resid = numpy.array(C - numpy.dot(A, bestPhaseOffsets)).ravel()
            resid = (C**2).sum(), (resid**2).sum()
            
            bestPhaseOffsets = list(bestPhaseOffsets)
            bestPhaseOffsets.insert(refAnt, 0.0)
            bestPhaseOffsets = numpy.array(bestPhaseOffsets)
            tempPhaseOffsets += bestPhaseOffsets
            
            valid = valid = numpy.where( numpy.abs(bestPhaseOffsets) < 1e6 )[0]
            metric = (numpy.abs(bestPhaseOffsets[valid])).max()
            if verbose:
                print('    ', metric)
            if metric < phaseCutoff:
                phaseOnly = False
                
            dataDict = _scaleData(dataDict, numpy.ones_like(bestPhaseOffsets), numpy.zeros_like(bestPhaseOffsets), bestPhaseOffsets)
            
        elif delayOnly:
            if verbose:
                print('  %iD' % (i+1,))
                
            A = _buildDelayOnlyA(aa, dataDict, simDict, chan, pol, refAnt=refAnt)
            C = _buildDelayOnlyC(aa, dataDict, simDict, chan, pol, refAnt=refAnt)
            
            good = numpy.where( numpy.isfinite(C) == 1 )[0]
            A = A[good,:]
            C = C[good]
            
            bestDelays, resid, rank, s = numpy.linalg.lstsq(A, C)
            resid = numpy.array(C - numpy.dot(A, bestDelays)).ravel()
            resid = (C**2).sum(), (resid**2).sum()
            
            bestDelays = list(bestDelays)
            bestDelays.insert(refAnt, 0.0)
            bestDelays = numpy.array(bestDelays)
            tempDelays += bestDelays
            
            valid = numpy.where( numpy.abs(bestDelays) < 1e6 )[0]
            metric = (numpy.abs(bestDelays[valid])).max()
            if verbose:
                print('    ', metric)
            if metric < delayCutoff:
                delayOnly = False
                
            dataDict = _scaleData(dataDict, numpy.ones_like(bestDelays), bestDelays, numpy.zeros_like(bestDelays))
            
        elif delayAndPhase:
            if verbose:
                print('  %iD+P' % (i+1,))
                
            A = _buildDelayAndPhaseA(aa, dataDict, simDict, chan, pol, refAnt=refAnt)
            C = _buildDelayAndPhaseC(aa, dataDict, simDict, chan, pol, refAnt=refAnt)
            
            good = numpy.where( numpy.isfinite(C) == 1 )[0]
            A = A[good,:]
            C = C[good]
            
            bestDelaysAndPhaseOffsets, resid, rank, s = numpy.linalg.lstsq(A, C)
            resid = numpy.array(C - numpy.dot(A, bestDelaysAndPhaseOffsets)).ravel()
            resid = (C**2).sum(), (resid**2).sum()
            
            bestDelays = list(bestDelaysAndPhaseOffsets[:(N-1)])
            bestDelays.insert(refAnt, 0.0)
            bestDelays = numpy.array(bestDelays)
            tempDelays += bestDelays
            
            bestPhaseOffsets = list(bestDelaysAndPhaseOffsets[(N-1):])
            bestPhaseOffsets.insert(refAnt, 0.0)
            bestPhaseOffsets = numpy.array(bestPhaseOffsets)
            tempPhaseOffsets += bestPhaseOffsets
            
            valid = numpy.where( numpy.abs(bestDelays) < 1e6 )[0]
            metric1 = (numpy.abs(bestDelays[valid])).max()
            metric2 = (numpy.abs(bestPhaseOffsets[valid])).max()
            if verbose:
                print('    ', metric1, metric2)
            if metric1 < delayCutoff and metric2 < phaseCutoff:
                delayAndPhase = False
                
            dataDict = _scaleData(dataDict, numpy.ones_like(bestDelays), bestDelays, bestPhaseOffsets)
            
        else:
            pass
            
    # Make sure the phase is (-pi, pi]
    tempPhaseOffsets %= 2*numpy.pi
    tempPhaseOffsets[numpy.where( tempPhaseOffsets >  numpy.pi )] -= 2*numpy.pi
    
    if verbose:
        print('Best Gains: ', tempGains)
    bestGains = tempGains
    
    if verbose:
        print('Best Delays: ', tempDelays)
    bestDelays = tempDelays
    
    if verbose:
        print('Best Phase Offsets: ', tempPhaseOffsets)
    bestPhaseOffsets = tempPhaseOffsets
    
    return dataDict, bestGains, bestDelays, bestPhaseOffsets


def phaseOnly(aa, dataDict, simDict, chan, pol, refAnt=0, nIter=30, amplitude=False, amplitudeCutoff=1.001, phaseCutoff=0.01, verbose=True):
    """
    Function to apply a phase-only (and, optionally, a amplitude) self-
    calibration to data stored in a readUVData dictionary and a model sky 
    stored in a lsl.sim.vis.buildSimSky dictionary for the given 
    polarization and channel(s).
    
    .. note::
        If the "amplitude" keyword is set to True, a three-element tuple of 
        self-cal'd data, gain coefficients, and phase offsets is returned 
        rather than the standard two-element tuple.
    """
    
    caldDict, gains, delays, phaseOffsets = _selfCal(aa, dataDict, simDict, chan, pol, refAnt=refAnt, nIter=nIter, 
                                            amplitude=amplitude, phaseOnly=True, 
                                            amplitudeCutoff=amplitudeCutoff, phaseCutoff=phaseCutoff, 
                                            verbose=verbose)
                                            
    if amplitude:
        return caldDict, gains, phaseOffsets
    else:
        return caldDict, phaseOffsets


def delayOnly(aa, dataDict, simDict, chan, pol, refAnt=0, nIter=30, amplitude=False, amplitudeCutoff=1.001, delayCutoff=0.2, verbose=True):
    """
    Function to apply a delay-only (and, optionally, a amplitude) self-
    calibration to data stored in a readUVData dictionary and a model sky 
    stored in a lsl.sim.vis.buildSimSky dictionary for the given 
    polarization and channel(s).
    
    .. note::
        If the "amplitude" keyword is set to True, a three-element tuple of 
        self-cal'd data, gain coefficients, and delays in ns is returned rather 
        than the standard two-element tuple.
    """
    
    caldDict, gains, delays, phaseOffsets = _selfCal(aa, dataDict, simDict, chan, pol, refAnt=refAnt, nIter=nIter, 
                                            amplitude=amplitude, delayOnly=True, 
                                            amplitudeCutoff=amplitudeCutoff, delayCutoff=delayCutoff,
                                            verbose=verbose)
                                            
    if amplitude:
        return caldDict, gains, delays
    else:
        return caldDict, delays


def delayAndPhase(aa, dataDict, simDict, chan, pol, refAnt=0, nIter=30, amplitude=False, amplitudeCutoff=1.001, delayCutoff=0.2, phaseCutoff=0.01, verbose=True):
    """
    Function to apply a delay and phase offset (and, optionally, a amplitude)
    self-calibration to data stored in a readUVData dictionary and a model 
    sky stored in a lsl.sim.vis.buildSimSky dictionary for the given 
    polarization and channel(s).
    
    .. note::
        If the "amplitude" keyword is set to True, a four-element tuple of 
        self-cal'd data, gain coefficients, delays in ns, and phase offsets 
        is returned rather than the standard three-element tuple.
    """
    
    caldDict, gains, delays, phaseOffsets = _selfCal(aa, dataDict, simDict, chan, pol, refAnt=refAnt, nIter=nIter, 
                                            amplitude=amplitude, delayAndPhase=True, 
                                            amplitudeCutoff=amplitudeCutoff, delayCutoff=delayCutoff, phaseCutoff=phaseCutoff, 
                                            verbose=verbose)
                                            
    if amplitude:
        return caldDict, gains, delays, phaseOffsets
    else:
        return caldDict, delays, phaseOffsets



# -*- coding: utf-8 -*-

# Python3 compatiability
from __future__ import print_function
import sys
if sys.version_info > (3,):
    xrange = range
    long = int
    import dbm
else:
    import anydbm as dbm

"""
Module that contains common values found in the MCS Joint Release 5 header file
src/exec/me.h and other functions useful for working with the MCS metadata.  
The header file values are:
 * ME_SSMIF_FORMAT_VERSION - SSMIF format version code
 * ME_MAX_NSTD - Maximum number of stands that can be described
 * ME_MAX_NFEE - Maximum number of FEEs that can be described
 * ME_MAX_FEEID_LENGTH - Number of characters in FEE ID name
 * ME_MAX_RACK - Maximum number of racks?
 * ME_MAX_PORT - Maximum number of ports?
 * ME_MAX_NRPD - Maxmimum number of RPD cables
 * ME_MAX_RPDID_LENGTH - Number of characters in the RPD ID name
 * ME_MAX_NSEP - Maximum number of SEP cable connections
 * ME_MAX_SEPID_LENGTH - Number of characters in the SEP ID name
 * ME_MAX_SEPCABL_LENGTH - Number of characters in the SEP cable ID name
 * ME_MAX_NARB - Maximum number of ARX boards
 * ME_MAX_NARBCH - Number of ARX channels per board
 * ME_MAX_ARBID_LENGTH - Number of characters in the ARX ID name
 * ME_MAX_NROACH - Maximum number of ROACH boards
 * ME_MAX_NROACHCH - Number of channels per ROACH board
 * ME_MAX_ROACHID_LENGTH - Number of characters in the ROACH board ID name
 * ME_MAX_NSERVER - Maximum number of servers
 * ME_MAX_SERVERID_LENGTH - Number of characters in the server ID name
 * ME_MAX_NDR - Maximum number of data recorders
 * ME_MAX_DRID_LENGTH - Number of characters in the DR ID name
 * ME_MAX_NPWRPORT - Maximum number of power ports
 * ME_MAX_SSNAME_LENGTH - Number of characters in the power port ID names, for 
                          codes used for PWR_NAME
 * LWA_MAX_NSTD - Maximum number of stands for the LWA
 * MIB_REC_TYPE_BRANCH - eType for MIB branch entries
 * MIB_REC_TYPE_VALUE - etype for MIB value entries
 * MIB_INDEX_FIELD_LENGTH - Number of characters in a MIB index field
 * MIB_LABEL_FIELD_LENGTH - Number of characters in a MIB label field
 * MIB_VAL_FIELD_LENGTH - Number of characters in a MIB value field
 * SSMIF_STRUCT - String representing the C structure of the binary SSMIF

The other functions:
 * Parse the binary packed metadata, 
"""

import re
import aipy
import math
import numpy
import ctypes
import struct
from datetime import datetime

from lsl.common import adp as adpCommon


__version__ = '0.1'
__revision__ = '$Rev: 2644 $'
__all__ = ['ME_SSMIF_FORMAT_VERSION', 'ME_MAX_NSTD', 'ME_MAX_NFEE', 'ME_MAX_FEEID_LENGTH', 'ME_MAX_RACK', 'ME_MAX_PORT', 
            'ME_MAX_NRPD', 'ME_MAX_RPDID_LENGTH', 'ME_MAX_NSEP', 'ME_MAX_SEPID_LENGTH', 'ME_MAX_SEPCABL_LENGTH', 
            'ME_MAX_NARB', 'ME_MAX_NARBCH', 'ME_MAX_ARBID_LENGTH', 'ME_MAX_NROACH', 'ME_MAX_NROACHCH', 'ME_MAX_ROACHID_LENGTH', 
            'ME_MAX_NSERVER', 'ME_MAX_SERVERID_LENGTH', 'ME_MAX_NDR', 'ME_MAX_DRID_LENGTH', 'ME_MAX_NPWRPORT', 
            'ME_MAX_SSNAME_LENGTH', 'LWA_MAX_NSTD', 'MIB_REC_TYPE_BRANCH', 'MIB_REC_TYPE_VALUE', 'MIB_INDEX_FIELD_LENGTH', 
            'MIB_LABEL_FIELD_LENGTH', 'MIB_VAL_FIELD_LENGTH', 'IS_32BIT_PYTHON', 
            'SSMIF_STRUCT', 'STATION_SETTINGS_STRUCT', 'SUBSYSTEM_STATUS_STRUCT', 'SUBSUBSYSTEM_STATUS_STRUCT', 
            'SSF_STRUCT', 'OSF_STRUCT', 'OSFS_STRUCT', 'BEAM_STRUCT', 'OSF2_STRUCT', 
            'delaytoMCSD', 'MCSDtodelay', 'gaintoMCSG', 'MCSGtogain',
            'mjdmpm2datetime', 'datetime2mjdmpm', 'status2string', 'summary2string', 'sid2string', 'cid2string', 
            'mode2string', 'parseCStruct', 'single2multi', 'applyPointingCorrection', 'MIB', 'MIBEntry', 
            '__version__', '__revision__', '__all__']


ME_SSMIF_FORMAT_VERSION = 8	# SSMIF format version code
ME_MAX_NSTD = 256			# Maximum number of stands that can be described
ME_MAX_NFEE = 256			# Maximum number of FEEs that can be described
ME_MAX_FEEID_LENGTH = 10		# Number of characters in FEE ID name
ME_MAX_RACK = 6			# Maximum number of racks?
ME_MAX_PORT = 50			# Maximum number of ports?
ME_MAX_NRPD = 512			# Maxmimum number of RPD cables
ME_MAX_RPDID_LENGTH = 25		# Number of characters in the RPD ID name
ME_MAX_NSEP = 512			# Maximum number of SEP cable connections
ME_MAX_SEPID_LENGTH = 25		# Number of characters in the SEP ID name
ME_MAX_SEPCABL_LENGTH = 25	# Number of characters in the SEP cable ID name
ME_MAX_NARB = 32			# Maximum number of ARX boards
ME_MAX_NARBCH = 16			# Number of ARX channels per board
ME_MAX_ARBID_LENGTH = 10		# Number of characters in the ARX ID name
ME_MAX_NROACH = 16			# Maximum number of ROACH boards
ME_MAX_NROACHCH = 32		# Number of channels per ROACH board
ME_MAX_ROACHID_LENGTH = 10	# Number of characters in the ROACH board ID name
ME_MAX_NSERVER = 7			# Maximum number of server
ME_MAX_SERVERID_LENGTH = 10	# Number of characters in the server ID name
ME_MAX_NDR = 2				# Maximum number of data recorders
ME_MAX_DRID_LENGTH = 10		# Number of characters in the DR ID name
ME_MAX_NPWRPORT = 50		# Maximum number of power ports
ME_MAX_SSNAME_LENGTH = 3		# Number of characters in the power port ID names, for codes used for PWR_NAME
LWA_MAX_NSTD = 256			# Maximum number of stands for the LWA
TPSS_FORMAT_VERSION = 6		# MCS0030 format version code
MIB_REC_TYPE_BRANCH = 0 		# eType for MIB branch entries
MIB_REC_TYPE_VALUE = 1 		# etype for MIB value entries
MIB_INDEX_FIELD_LENGTH = 12	# Number of characters in a MIB index field
MIB_LABEL_FIELD_LENGTH = 32	# Number of characters in a MIB label field
MIB_VAL_FIELD_LENGTH = 8192	# Number of characters in a MIB value field


IS_32BIT_PYTHON = True if struct.calcsize("P") == 4 else False


SSMIF_STRUCT = """
    int    iFormatVersion;           /* FORMAT_VERSION */
    char   sStationID[3];            /* STATION_ID */
    double fGeoN;                    /* GEO_N */
    double fGeoE;                    /* GEO_E */
    double fGeoEl;                   /* GEO_EL */
    int    nStd;                     /* N_STD */
    double fStdLx[ME_MAX_NSTD];      /* STD_LX[] */
    double fStdLy[ME_MAX_NSTD];      /* STD_LY[] */
    double fStdLz[ME_MAX_NSTD];      /* STD_LZ[] */
    int    iAntStd[2*ME_MAX_NSTD];   /* ANT_STD[] */
    int    iAntOrie[2*ME_MAX_NSTD];  /* ANT_ORIE[] */
    int    iAntStat[2*ME_MAX_NSTD];  /* ANT_STAT[] */
    float  fAntTheta[2*ME_MAX_NSTD]; /* ANT_THETA[] */
    float  fAntPhi[2*ME_MAX_NSTD];   /* ANT_PHI[] */
    int    eAntDesi[2*ME_MAX_NSTD];  /* ANT_DESI[] */
    %s
    int    nFEE;                     /* N_FEE */
    char   sFEEID[ME_MAX_NFEE][ME_MAX_FEEID_LENGTH+1]; /* FEE_ID[] */
    int    iFEEStat[ME_MAX_NFEE];    /* FEE_STAT[] */
    int    eFEEDesi[ME_MAX_NFEE];    /* FEE_DESI[] */
    float  fFEEGai1[ME_MAX_NFEE];    /* FEE_GAI1[] */
    float  fFEEGai2[ME_MAX_NFEE];    /* FEE_GAI2[] */
    int    iFEEAnt1[ME_MAX_NFEE];    /* FEE_ANT1[] */
    int    iFEEAnt2[ME_MAX_NFEE];    /* FEE_ANT2[] */
    int    iFEERack[ME_MAX_NFEE];    /* FEE_RACK[] */
    int    iFEEPort[ME_MAX_NFEE];    /* FEE_PORT[] */
    int    nRPD;                     /* N_RPD */
    char   sRPDID[ME_MAX_NRPD][ME_MAX_RPDID_LENGTH+1]; /* RPD_ID[] */
    int    iRPDStat[ME_MAX_NRPD];    /* RPD_STAT[] */
    int    eRPDDesi[ME_MAX_NRPD];    /* RPD_DESI[] */
    float  fRPDLeng[ME_MAX_NRPD];    /* RPD_LENG[] */
    float  fRPDVF[ME_MAX_NRPD];      /* RPD_VF[] */
    float  fRPDDD[ME_MAX_NRPD];      /* RPD_DD[] */
    float  fRPDA0[ME_MAX_NRPD];      /* RPD_A0[] */
    float  fRPDA1[ME_MAX_NRPD];      /* RPD_A1[] */
    float  fRPDFref[ME_MAX_NRPD];    /* RPD_FREF[] */
    float  fRPDStr[ME_MAX_NRPD];     /* RPD_STR[] */
    int    iRPDAnt[ME_MAX_NRPD];     /* RPD_ANT[] */
    int    nSEP;                     /* N_SEP */
    char   sSEPID[ME_MAX_NSEP][ME_MAX_SEPID_LENGTH+1]; /* SEP_ID[] */
    int    iSEPStat[ME_MAX_NSEP];    /* SEP_STAT[] */
    char   sSEPCabl[ME_MAX_NSEP][ME_MAX_SEPCABL_LENGTH+1]; /* SEP_Cabl[] */
    float  fSEPLeng[ME_MAX_NSEP];    /* SEP_LENG[] */
    int    eSEPDesi[ME_MAX_NSEP];    /* SEP_DESI[] */
    float  fSEPGain[ME_MAX_NSEP];    /* SEP_GAIN[] */
    int    iSEPAnt[ME_MAX_NSEP];     /* SEP_ANT[] */
    int    nARB;                     /* N_ARB */
    int    nARBCH;                   /* N_ARBCH */
    char   sARBID[ME_MAX_NARB][ME_MAX_ARBID_LENGTH+1]; /* ARB_ID[] */
    int    iARBSlot[ME_MAX_NARB];    /* ARB_SLOT[] */
    int    eARBDesi[ME_MAX_NARB];    /* ARB_DESI[] */
    int    iARBRack[ME_MAX_NARB];    /* ARB_RACK[] */
    int    iARBPort[ME_MAX_NARB];    /* ARB_PORT[] */
    int    eARBStat[ME_MAX_NARB][ME_MAX_NARBCH];       /* ARB_STAT[][] */
    float  fARBGain[ME_MAX_NARB][ME_MAX_NARBCH];        /* ARB_GAIN[][] */
    int    iARBAnt[ME_MAX_NARB][ME_MAX_NARBCH];        /* ARB_ANT[][] */
    char   sARBIN[ME_MAX_NARB][ME_MAX_NARBCH][ME_MAX_ARBID_LENGTH+1]; /* ARB_IN[][] */
    char   sARBOUT[ME_MAX_NARB][ME_MAX_NARBCH][ME_MAX_ARBID_LENGTH+1]; /* ARB_OUT[][] */
    int    nRoach;                     /* N_ROACH */
    int    nRoachCh;                   /* N_ROACHCH */
    char   sRoachID[ME_MAX_NROACH][ME_MAX_ROACHID_LENGTH+1]; /* ROACH_ID[] */
    char   sRoachSlot[ME_MAX_NROACH][ME_MAX_ROACHID_LENGTH+1]; /* ROACH_SLOT[] */
    int    eRoachDesi[ME_MAX_NROACH]; /* ROACH_DESI[] */
    int    eRoachStat[ME_MAX_NROACH][ME_MAX_NROACHCH];       /* ROACH_STAT[][] */
    char   sRoachINR[ME_MAX_NROACH][ME_MAX_NROACHCH][ME_MAX_ROACHID_LENGTH+1]; /* ROACH_INR[][] */
    char   sRoachINC[ME_MAX_NROACH][ME_MAX_NROACHCH][ME_MAX_ROACHID_LENGTH+1]; /* ROACH_INC[][] */
    int    iRoachAnt[ME_MAX_NROACH][ME_MAX_NROACHCH];        /* ROACH_ANT[][] */
    int    nServer;                     /* N_SERVER */
    char   sServerID[ME_MAX_NSERVER][ME_MAX_SERVERID_LENGTH+1]; /* SERVER_ID[] */
    char   sServerSlot[ME_MAX_NSERVER][ME_MAX_SERVERID_LENGTH+1]; /* SERVER_SLOT[] */
    int    eServerStat[ME_MAX_NSERVER];       /* SERVER_STAT[] */
    int    eServerDesi[ME_MAX_NSERVER];       /* SERVER_DESI[] */
    int    nDR;                     /* N_DR */
    int    eDRStat[ME_MAX_NDR];       /* DR_STAT[] */
    char   sDRID[ME_MAX_NDR][ME_MAX_DRID_LENGTH+1]; /* DR_ID[] */
    char   sDRPC[ME_MAX_NDR][ME_MAX_DRID_LENGTH+1]; /* DR_PC[] */
    int    iDRDP[ME_MAX_NDR];       /* DR_DP[] */
    int    nPwrRack;                /* N_PWR_RACK */
    int    nPwrPort[ME_MAX_RACK];   /* N_PWR_PORT[] */
    int    ePwrSS[ME_MAX_RACK][ME_MAX_NPWRPORT]; /* PWR_SS[][], converted to a LWA_SID_ value */
    char   sPwrName[ME_MAX_RACK][ME_MAX_NPWRPORT][ME_MAX_SSNAME_LENGTH+1]; /* PWR_NAME[][] */
    int    eCRA;                /* MCS_CRA */
    float  fPCAxisTh; /* PC_AXIS_TH */
    float  fPCAxisPh; /* PC_AXIS_PH */
    float  fPCRot;    /* PC_ROT */
""" % ("short int junk;\n" if IS_32BIT_PYTHON else "")


STATION_SETTINGS_STRUCT = """
    signed short int mrp_asp; // SESSION_MRP_ASP // MRP_ASP
    signed short int mrp_dp;  // SESSION_MRP_DP_ // MRP_DP_
    signed short int mrp_dr1; // SESSION_MRP_DR1 // MRP_DR1
    signed short int mrp_dr2; // SESSION_MRP_DR2 // MRP_DR2
    signed short int mrp_dr3; // SESSION_MRP_DR3 // MRP_DR3
    signed short int mrp_dr4; // SESSION_MRP_DR4 // MRP_DR4
    signed short int mrp_dr5; // SESSION_MRP_DR5 // MRP_DR5
    signed short int mrp_shl; // SESSION_MRP_SHL // MRP_SHL
    signed short int mrp_mcs; // SESSION_MRP_MCS // MRP_MCS
    signed short int mup_asp; // SESSION_MUP_ASP // MUP_ASP
    signed short int mup_dp;  // SESSION_MUP_DP_ // MUP_DP_
    signed short int mup_dr1; // SESSION_MUP_DR1 // MUP_DR1
    signed short int mup_dr2; // SESSION_MUP_DR2 // MUP_DR2
    signed short int mup_dr3; // SESSION_MUP_DR3 // MUP_DR3
    signed short int mup_dr4; // SESSION_MUP_DR4 // MUP_DR4
    signed short int mup_dr5; // SESSION_MUP_DR5 // MUP_DR5
    signed short int mup_shl; // SESSION_MUP_SHL // MUP_SHL
    signed short int mup_mcs; // SESSION_MUP_MCS // MUP_MCS
    signed short int fee[LWA_MAX_NSTD];     // OBS_FEE[LWA_MAX_NSTD][2]  // FEE[LWA_MAX_NSTD]
    signed short int asp_flt[LWA_MAX_NSTD]; // OBS_ASP_FLT[LWA_MAX_NSTD] // ASP_FLT[LWA_MAX_NSTD]
    signed short int asp_at1[LWA_MAX_NSTD]; // OBS_ASP_AT1[LWA_MAX_NSTD] // ASP_AT1[LWA_MAX_NSTD]
    signed short int asp_at2[LWA_MAX_NSTD]; // OBS_ASP_AT2[LWA_MAX_NSTD] // ASP_AT2[LWA_MAX_NSTD]
    signed short int asp_ats[LWA_MAX_NSTD]; // OBS_ASP_ATS[LWA_MAX_NSTD] // ASP_ATS[LWA_MAX_NSTD]
    signed short int tbf_gain; // OBS_TBF_GAIN // TBF_GAIN
    signed short int tbn_gain; // OBS_TBN_GAIN // TBN_GAIN
    signed short int drx_gain; // OBS_DRX_GAIN // DRX_GAIN
"""


SUBSYSTEM_STATUS_STRUCT = """
    int summary;
    %s
    char info[256];
    long tv[2];
""" % ("short int junk;\n" if IS_32BIT_PYTHON else "",)


SUBSUBSYSTEM_STATUS_STRUCT = """
    int    eFEEStat[ME_MAX_NFEE];                      /* FEE_STAT[] */
    int    eRPDStat[ME_MAX_NRPD];                      /* RPD_STAT[] */
    int    eSEPStat[ME_MAX_NSEP];                      /* SEP_STAT[] */
    int    eARBStat[ME_MAX_NARB][ME_MAX_NARBCH];       /* ARB_STAT[][] */
    int    eRoachStat[ME_MAX_NROACH][ME_MAX_NROACHCH]; /* ROACH_STAT[][] */
    int    eServerStat[ME_MAX_NSERVER];                /* SERVER_STAT[] */
    int    eDRStat[ME_MAX_NDR];                        /* DR_STAT[] */
"""


SSF_STRUCT = """
    unsigned short int FORMAT_VERSION;
    char PROJECT_ID[9];
    unsigned int SESSION_ID;
    unsigned short int SESSION_CRA;
    signed short int SESSION_DRX_BEAM;
    char SESSION_SPC[32];
    %s
    unsigned long int SESSION_START_MJD;
    unsigned long int SESSION_START_MPM;
    unsigned long int SESSION_DUR;
    unsigned int SESSION_NOBS;
    signed short int SESSION_MRP_ASP;
    signed short int SESSION_MRP_DP_;
    signed short int SESSION_MRP_DR1;
    signed short int SESSION_MRP_DR2;
    signed short int SESSION_MRP_DR3;
    signed short int SESSION_MRP_DR4;
    signed short int SESSION_MRP_DR5;
    signed short int SESSION_MRP_SHL;
    signed short int SESSION_MRP_MCS;
    signed short int SESSION_MUP_ASP;
    signed short int SESSION_MUP_DP_;
    signed short int SESSION_MUP_DR1;
    signed short int SESSION_MUP_DR2;
    signed short int SESSION_MUP_DR3;
    signed short int SESSION_MUP_DR4;
    signed short int SESSION_MUP_DR5;
    signed short int SESSION_MUP_SHL;
    signed short int SESSION_MUP_MCS;
    signed char SESSION_LOG_SCH;
    signed char SESSION_LOG_EXE;
    signed char SESSION_INC_SMIB;
    signed char SESSION_INC_DES;
""" % ("short int junk;\n" if IS_32BIT_PYTHON else "",)


OSF_STRUCT = """
    unsigned short int FORMAT_VERSION;
    char               PROJECT_ID[9];
    unsigned int       SESSION_ID;
    signed short int   SESSION_DRX_BEAM;
    char               SESSION_SPC[32];
    unsigned int       OBS_ID; 
    unsigned long int  OBS_START_MJD;
    unsigned long int  OBS_START_MPM;
    unsigned long int  OBS_DUR;
    unsigned short int OBS_MODE;
    char               OBS_BDM[32];  /* added 140310 */
    float              OBS_RA;
    float              OBS_DEC;
    unsigned short int OBS_B;
    unsigned int       OBS_FREQ1;
    unsigned int       OBS_FREQ2;
    unsigned short int OBS_BW;
    unsigned int       OBS_STP_N;
    unsigned short int OBS_STP_RADEC;
"""


OSFS_STRUCT = """
    float              OBS_STP_C1;
    float              OBS_STP_C2;
    unsigned int       OBS_STP_T;
    unsigned int       OBS_STP_FREQ1;
    unsigned int       OBS_STP_FREQ2;
    unsigned short int OBS_STP_B;
"""


BEAM_STRUCT = """
    unsigned short int OBS_BEAM_DELAY[2*LWA_MAX_NSTD];
    signed short int   OBS_BEAM_GAIN[LWA_MAX_NSTD][2][2];
"""


OSF2_STRUCT = """
    signed short int   OBS_FEE[LWA_MAX_NSTD][2];
    signed short int   OBS_ASP_FLT[LWA_MAX_NSTD];
    signed short int   OBS_ASP_AT1[LWA_MAX_NSTD];
    signed short int   OBS_ASP_AT2[LWA_MAX_NSTD];
    signed short int   OBS_ASP_ATS[LWA_MAX_NSTD];
    unsigned int       OBS_TBF_SAMPLES;
    signed short int   OBS_TBF_GAIN;
    signed short int   OBS_TBN_GAIN;
    signed short int   OBS_DRX_GAIN;
    unsigned int       alignment;
"""


_cDecRE = re.compile(r'(?P<type>[a-z][a-z \t]+)[ \t]+(?P<name>[a-zA-Z_0-9]+)(\[(?P<d1>[\*\+A-Z_\d]+)\])?(\[(?P<d2>[\*\+A-Z_\d]+)\])?(\[(?P<d3>[\*\+A-Z_\d]+)\])?(\[(?P<d4>[\*\+A-Z_\d]+)\])?;')


def parseCStruct(cStruct, charMode='str', endianness='native'):
    """
    Function to take a C structure declaration and build a ctypes.Structure out 
    of it with the appropriate alignment, character interpretation*, and endianness
    (little, big, network, or native).
    
    .. note::  ctypes converts character arrays to Python strings until the first null is
    incountered.  This behavior causes problems for multi-dimension arrays of null
    filled strings.  By setting charMode to 'int', all char types are retuned as 
    bytes which can be converted to strings via chr().
    """
    
    # Figure out how to deal with character arrays
    if charMode not in ('str', 'int'):
        raise RuntimeError("Unknown character mode: '%s'" % charMode)
    if charMode == 'str':
        baseCharType = ctypes.c_char
    else:
        baseCharType = ctypes.c_byte
    
    # Hold the basic fields and dimensions
    fields = []
    dims2 = {}
    
    # Split into lines and go!
    cStruct = cStruct.split('\n')
    for line in cStruct:
        ## Skip structure declaration, blank lines, comments, and lines too short to hold a 
        ## declaration
        line = line.strip().rstrip()
        if '{' in line or '}' in line:
            continue
        if line[:2] == '//':
            continue
        if len(line) < 5:
            continue
            
        ## RegEx the line to find the type, name, and dimensions (if needed) for
        ## the next structure variable
        mtch = _cDecRE.search(line)
        if mtch is None:
            raise RuntimeError("Unparseable line: '%s'" % line)
        
        dec = mtch.group('type')
        dec = dec.rstrip()
        name = mtch.group('name')
        
        try:
            d1 = mtch.group('d1')
            if d1 is not None:
                d1 = eval(d1)
            d2 = mtch.group('d2')
            if d2 is not None:
                d2 = eval(d2)
            d3 = mtch.group('d3')
            if d3 is not None:
                d3 = eval(d3)
            d4 = mtch.group('d4')
            if d4 is not None:
                d4 = eval(d4)
        except NameError:
            raise RuntimeError("Unknown value in array index: '%s'" % line)
        
        ## Basic data types
        if dec in ('signed int', 'int'):
            typ = ctypes.c_int
        elif dec == 'unsigned int':
            typ = ctypes.c_uint
        elif dec in ('signed short int', 'signed short', 'short int', 'short'):
            typ = ctypes.c_short
        elif dec in ('unsigned short int', 'unsigned short'):
            typ = ctypes.c_ushort
        elif dec in ('signed long int', 'signed long', 'long int', 'long'):
            if IS_32BIT_PYTHON:
                typ = ctypes.c_longlong
            else:
                typ = ctypes.c_long
        elif dec in ('unsigned long int', 'unsigned long'):
            if IS_32BIT_PYTHON:
                typ = ctypes.c_uint64
            else:
                typ = ctypes.c_ulong
        elif dec in ('signed long long', 'long long'):
            typ = ctypes.c_longlong
        elif dec == 'unsigned long long':
            typ = ctypes.c_uint64
        elif dec == 'float':
            typ = ctypes.c_float
        elif dec == 'double':
            typ = ctypes.c_double
        elif dec == 'char':
            typ = baseCharType
        elif dec == 'signed char':
            typ = ctypes.c_byte
        elif dec == 'unsigned char':
            typ = ctypes.c_ubyte
        else:
            raise RuntimeError("Unparseable line: '%s' -> type: %s, name: %s, dims: %s, %s, %s %s" % (line, dec, name, d1, d2, d3, d4))
        
        ## Array identification and construction
        dims2[name] = []
        if d1 is not None:
            count = d1
            dims2[name].append(d1)
            
            if d2 is not None:
                count *= d2
                dims2[name].append(d2)
            if d3 is not None:
                count *= d3
                dims2[name].append(d3)
            if d4 is not None:
                count *= d4
                dims2[name].append(d4)
                
            typ *= count
        
        ## Append
        fields.append( (name, typ) )
    
    # ctypes creation - endianess
    endianness = endianness.lower()
    if endianness not in ('little', 'big', 'network', 'native'):
        raise RuntimeError("Unknown endianness: '%s'" % endianness)
    
    if endianness == 'little':
        endianness = ctypes.LittleEndianStructure
    elif endianness == 'big':
        endianness = ctypes.BigEndianStructure
    elif endianness == 'network':
        endianness = ctypes.BigEndianStructure
    else:
        endianness = ctypes.Structure
    
    # ctypes creation - actual
    class MyStruct(endianness):
        """
        ctypes.Structure of the correct endianness for the provided
        C structure.  
        
        In addition to the standard attributes defined for a ctypes.Structure 
        instance there are a few additional attributes related to the parsed C
        structure.  They are:
         * origC - String containing the original C structure
         * dims  - Dictionary of the dimensionallity of the data, if needed, 
                   keyed by variable name
        """
        
        origC = '\n'.join(cStruct)
        
        _fields_ = fields
        _pack_ = 8	# Pack it like we are 64-bit
        dims = dims2
        
        def __str__(self):
            """
            Print out the structure in a nice easy-to-read formation that
            captures the various structure elements, their data types, and 
            their values.
            """
            
            out = ''
            for f,d in self._fields_:
                out += '%s (%s): %s\n' % (f, d, eval("self.%s" % f))
            return out
            
        def sizeof(self):
            """
            Return the size, in bytes, of the structure.
            """
            
            return ctypes.sizeof(self)
            
        def returnDict(self):
            """
            Return the structure as a simple Python dictionary keyed off the
            structure elements.
            """
            
            output = {}
            for f,d in self._fields_:
                output[f] = eval("self.%s" % f)
            return output
    
    # Create and return
    return MyStruct()


def _twoByteSwap(value):
    return ((value & 0xFF) << 8) | ((value >> 8) & 0xFF)


def delaytoMCSD(delay):
    """
    Given a delay in ns, convert it to a course and fine portion and into 
    the form expected by MCS in a custom beamforming SDF (little endian 
    16.12 unsigned integer).

    .. versionadded:: 0.6.3
    """
    
    return _twoByteSwap( adpCommon.delaytoDPD(delay) )


def MCSDtodelay(delay):
    """
    Given delay value from an OBS_BEAM_DELAY field in a custom beamforming 
    SDF, return the delay in ns.

    .. versionadded:: 0.6.3
    """
    
    return adpCommon.DPDtodelay( _twoByteSwap(delay) )


def gaintoMCSG(gain):
    """
    Given a gain (between 0 and 1), convert it to a gain in the form 
    expected by MCS in a custom beamforming SDF (little endian 16.1 
    signed integer).

    .. versionadded::0.6.3
    """
    
    return _twoByteSwap( adpCommon.gaintoDPG(gain) )


def MCSGtogain(gain):
    """
    Given a gain value from an OBS_BEAM_GAIN field in a custom beamforming
    SDF, return the decimal equivalent.

    .. versionadded:: 0.6.3
    """
    
    return adpCommon.DPGtogain( _twoByteSwap(gain) )


def mjdmpm2datetime(mjd, mpm):
    """
    Convert a MJD, MPM pair to a UTC-aware datetime instance.
    
    .. versionadded:: 0.5.2
    """
    
    unix = mjd*86400.0 + mpm/1000.0 - 3506716800.0
    return datetime.utcfromtimestamp(unix)


def datetime2mjdmpm(dt):
    """
    Convert a UTC datetime instance to a MJD, MPM pair (returned as a 
    two-element tuple).
    
    Based off: http://paste.lisp.org/display/73536
    
    .. versionadded:: 0.5.2
    """
    
    year        = dt.year             
    month       = dt.month      
    day         = dt.day    

    hour        = dt.hour
    minute      = dt.minute
    second      = dt.second     
    millisecond = dt.microsecond / 1000

    # compute MJD         
    # adapted from http://paste.lisp.org/display/73536
    # can check result using http://www.csgnetwork.com/julianmodifdateconv.html
    a = (14 - month) // 12
    y = year + 4800 - a          
    m = month + (12 * a) - 3                    
    p = day + (((153 * m) + 2) // 5) + (365 * y)   
    q = (y // 4) - (y // 100) + (y // 400) - 32045
    mjdi = int(math.floor( (p+q) - 2400000.5))
    mjd = mjdi

    # compute MPM
    mpmi = int(math.floor( (hour*3600 + minute*60 + second)*1000 + millisecond ))
    mpm = mpmi
    return (mjd, mpm)


def status2string(code):
    """
    Convert a numerical MCS status code to a string.
    """
    
    # Make sure we have an integer
    code = int(code)
    
    # Loop through the options
    if code == 0:
        return "Not installed"
    elif code == 1:
        return "Bad"
    elif code == 2:
        return "Suspect, possibly bad"
    elif code == 3:
        return "OK"
    else:
        return "Unknown status code '%i'" % code


def summary2string(code):
    """
    Convert a numerical MCS overall status code to an explination.
    """
    
    if code < 0 or code > 6:
        raise ValueError("Invalid code %i" % code)
    
    if code == 0:
        return "Not normally used"
    elif code == 1:
        return "Normal"
    elif code == 2:
        return "Warning - issue(s) found, but still fully operational"
    elif code == 3:
        return "Error - problems found which limit or prevent proper function"
    elif code == 4:
        return "Booting - initializing; not yet fully operational"
    elif code == 5:
        return "Shutdown - shutting down; not ready for operation"
    else:
        return "Status is unknown"


def sid2string(sid):
    """
    Convert a MCS subsystem ID code into a string.
    """
    
    if sid < 1 or sid > 19:
        raise ValueError("Invalid sid code %i" % sid)
    
    if sid < 9:
        return "Null subsystem #%i" % sid
    elif sid == 10:
        return "MCS"
    elif sid == 11:
        return "SHL"
    elif sid == 12:
        return "ASP"
    elif sid == 14:
        return "DR #1"
    elif sid == 15:
        return "DR #2"
    elif sid == 16:
        return "DR #3"
    elif sid == 17:
        return "DR #4"
    elif sid == 18:
        return "DR #5"
    elif sid == 19:
        return "ADP"
    else:
        raise ValueError("Invalid sid code %i" % sid)


def cid2string(cid):
    """
    Convert a MCS command code into a string.
    """
    
    if cid < 0 or cid > 41:
        raise ValueError("Invalid cid code %i" % cid)
    
    if cid == 0:
        return "MCSSHT"
    elif cid == 1:
        return "PNG"
    elif cid == 2:
        return "RPT"
    elif cid == 3:
        return "SHT"
    elif cid == 4:
        return "INI"
    elif cid == 5:
        return "TMP"
    elif cid == 6:
        return "DIF"
    elif cid == 7:
        return "PWR"
    elif cid == 8:
        return "FIL"
    elif cid == 9:
        return "AT1"
    elif cid == 10:
        return "AT2"
    elif cid == 11:
        return "ATS"
    elif cid == 12:
        return "FPW"
    elif cid == 13:
        return "RXP"
    elif cid == 14:
        return "FEP"
    elif cid == 16:
        return "TBN"
    elif cid == 17:
        return "DRX"
    elif cid == 18:
        return "BAM"
    elif cid == 19:
        return "FST"
    elif cid == 20:
        return "CLK"
    elif cid == 21:
        return "REC"
    elif cid == 22:
        return "DEL"
    elif cid == 23:
        return "STP"
    elif cid == 24:
        return "GET"
    elif cid == 25:
        return "CPY"
    elif cid == 26:
        return "DMP"
    elif cid == 27:
        return "FMT"
    elif cid == 28:
        return "DWN"
    elif cid == 29:
        return "UP_"
    elif cid == 30:
        return "SEL"
    elif cid == 31:
        return "SYN"
    elif cid == 32:
        return "TST"
    elif cid == 33:
        return "BUF"
    elif cid == 34:
        return "NUL"
    elif cid == 35:
        return "ESN"
    elif cid == 36:
        return "ESF"
    elif cid == 37:
        return "OBS"
    elif cid == 38:
        return "OBE"
    elif cid == 39:
        return "SPC"
    elif cid == 40:
        return "TBF"
    elif cid == 41:
        return "COR"
    else:
        raise ValueError("Invalid cid code %i" % cid)


def mode2string(mode):
    """
    Convert a MCS numeric observing mode into a string.
    """
    
    if mode < 1 or mode > 8:
        raise ValueError("Invalid observing mode %i" % mode)
    
    if mode == 1:
        return "TRK_RADEC"
    elif mode == 2:
        return "TRK_SOL"
    elif mode == 3:
        return "TRK_JOV"
    elif mode == 4:
        return "STEPPED"
    elif mode == 6:
        return "TBN"
    elif mode == 7:
        return "DIAG1"
    elif mode == 8:
        return "TBF"
    else:
        raise ValueError("Invalid observing mode %i" % mode)


def single2multi(inputList, dim1, dim2=None, dim3=None, dim4=None):
    if dim4 is not None:
        return _single2four(inputList, dim1, dim2, dim3, dim4)
        
    elif dim3 is not None:
        return _single2three(inputList, dim1, dim2, dim3)

    elif dim2 is not None:
        return _single2two(inputList, dim1, dim2)
        
    else:
        return list(inputList)


def _single2two(inputList, dim1, dim2):
    """
    Convert a flatten list into a two-dimensional list.  This is useful
    for converting flat lists of ctypes into their two-dimensional forms.
    """
    
    if dim1*dim2 < len(inputList):
        raise ValueError("Incompatiable dimensions: input=%i, output=%i by %i" % (len(inputList), dim1, dim2))
    
    outputList = []
    for i in xrange(dim1):
        outputList.append( [None for k in xrange(dim2)] )
        for j in xrange(dim2):
            try:
                outputList[i][j] = inputList[dim2*i+j]
            except IndexError:
                pass
    
    return outputList


def _single2three(inputList, dim1, dim2, dim3):
    """
    Convert a flatten list into a three-dimensional list.  This is useful
    for converting flat lists of ctypes into their three-dimensional forms.
    """
    
    if dim1*dim2*dim3 < len(inputList):
        raise ValueError("Incompatiable dimensions: input=%i, output=%i by %i by %i" % (len(inputList), dim1, dim2, dim3))
    
    outputList = []
    for i in xrange(dim1):
        outputList.append( [[None for l in xrange(dim3)] for k in xrange(dim2)] )
        for j in xrange(dim2):
            for k in xrange(dim3):
                try:
                    outputList[i][j][k] = inputList[dim2*dim3*i+dim3*j+k]
                except IndexError:
                    pass
    
    return outputList


def _single2four(inputList, dim1, dim2, dim3, dim4):
    """
    Convert a flatten list into a four-dimensional list.  This is useful
    for converting flat lists of ctypes into their four-dimensional forms.
    """
    
    if dim1*dim2*dim3*dim4 < len(inputList):
        raise ValueError("Incompatiable dimensions: input=%i, output=%i by %i by %i by %i" % (len(inputList), dim1, dim2, dim3, dim4))
    
    outputList = []
    for i in xrange(dim1):
        outputList.append( [[[None for m in xrange(dim4)] for l in xrange(dim3)] for k in xrange(dim2)] )
        for j in xrange(dim2):
            for k in xrange(dim3):
                for l in xrange(dim4):
                    try:
                        outputList[i][j][k][l] = inputList[dim2*dim3*dim4*i+dim3*dim4*j+dim4*k+l]
                    except IndexError:
                        pass
    
    return outputList


def _getRotationMatrix(theta, phi, psi, degrees=True):
    """
    Generate the rotation matrix for a rotation about a specified axis.
    
    http://en.wikipedia.org/wiki/Rotation_matrix#Rotation_matrix_from_axis_and_angle
    """
    
    if degrees:
        theta = theta * numpy.pi/180.0
        phi   = phi * numpy.pi/180.0
        psi   = psi * numpy.pi/180.0
        
    # Axis
    u = numpy.array([numpy.cos(phi)*numpy.sin(theta), numpy.sin(phi)*numpy.sin(theta), numpy.cos(theta)])
    
    # Rotation matrix
    rot  = numpy.eye(3)*numpy.cos(psi) 
    rot += numpy.sin(psi)*numpy.array([[0, -u[2], u[1]], [u[2], 0, -u[0]], [-u[1], u[0], 0]]) 
    rot += (1-numpy.cos(psi))*numpy.tensordot(u, u, axes=0)
    
    return rot


def applyPointingCorrection(az, el, theta, phi, psi, lat=34.070, degrees=True):
    """
    Given a azimuth and elevation pair, and an axis to rotate about, 
    perform the rotation.
    
    http://en.wikipedia.org/wiki/Rotation_matrix#Rotation_matrix_from_axis_and_angle
    """
    
    # Get the rotation matrix
    rot = _getRotationMatrix(theta, phi, psi, degrees=degrees)
    
    # Convert the az,alt coordinates to the unit vector
    if degrees:
        xyz = aipy.coord.azalt2top((az*numpy.pi/180.0, el*numpy.pi/180.0))
    else:
        xyz = aipy.coord.azalt2top((az, el))
        
    # Rotate
    xyzP = numpy.dot(rot, xyz)
    
    azP, elP = aipy.coord.top2azalt(xyzP)
    if degrees:
        azP *= 180/numpy.pi
        elP *= 180/numpy.pi
        
    return azP, elP


class MIB(object):
    """
    Class to represent an entire MCS MIB.
    
    .. versionadded:: 0.6.5
    """
    
    def __init__(self):
        self.entries = {}
        self.mapper = {}
        self.invMapper = {}
        
    def __str__(self):
        """
        Describe the MIB as a string.
        """
        
        nEntry = len(self.entries.keys())
        times = [self.entries[k].updateTime for k in self.entries.keys()]
        return "MIB containing %i entries from %s to %s" % (nEntry, min(times), max(times))
        
    def __getitem__(self, key):
        """
        Allow access to the MIB through either the index of name.
        """
        
        try:
            return self.entries[key]
        except KeyError:
            return self.entries[self.invMapper[key]]
            
    def keys(self, name=False):
        """
        Return a list of entry indicies (or names if the 'name' keyword is set 
        to True) for the MIB.
        """
        
        if name:
            # Index -> Name
            output = []
            for k in self.entries.keys():
                try:
                    output.append( self.mapper[k] )
                except KeyError:
                    output.append(k)
            return output
        else:
            # Index
            return self.entries.keys()
            
    def parseInitFile(self, filename):
        """
        Given a MCS MIB initialization file, i.e., ASP_MIB_init.dat, create a 
        dictionary that maps indicies to human-readable names that can be used
        to clarify a MIBEntry.  Return this dictionary.
        """
        
        # Open the file
        with open(filename, 'r') as fh:
            # Go!
            self.mapper = {}
            self.invMapper = {}
            for line in fh:
                ## Parse out the line
                line = line.replace('\n', '')
                eType, index, name, default, dbmType, icdType = line.split(None, 5)
                
                ## Skip over branches
                if eType == 'B':
                    continue
                    
                ## Remember the mapping
                self.mapper[index] = name
                self.invMapper[name] = index
                
        # Done
        return True
        
    def fromFile(self, filename, initFilename=None):
        """
        Given the name of a GDBM database file, initialize the MIB.  
        Optionally, use the name of the MCS MIB initialization file to
        help convert indicies to names.
        """
        
        # Parse the init. file (if we have one)
        if initFilename is not None:
            self.parseInitFile(initFilename)
            
        # Make sure we have the .pag file
        if filename[-4:] == '.dir':
            filename = filename.replace('.dir', '.pag')
            
        # Open the database
        db = dbm.open(filename, 'ru')
        
        # Go!
        entry = db.firstkey()
        while entry is not None:
            value = db[entry]
            
            try:
                record = MIBEntry()
                record.fromEntry(value)
                self.entries[record.index] = record
                
            except ValueError:
                pass
                
            entry = db.nextkey(entry)
        db.close()
        
        # Done
        return True


class MIBEntry(object):
    """
    Class for accessing and storing MCS MIB information contained in a GDBM 
    database.
    
    .. versionadded:: 0.6.5
    """
    
    def __init__(self):
        """
        Create the MIBEntry instance and fill with dummy values.
        """
        
        # Basic information straight out of the mcs.h/record structure
        self.eType = 0
        self.index = ''
        self.value = ''
        self.dbmType = ''
        self.icdType = ''
        self._tv = (0, 0)
        
        # Additional information determined from the basic information
        self.updateTime = datetime.utcfromtimestamp(0)
        
    def __str__(self):
        """
        Represent the class as a string for display purposes.
        """
        
        return "Index: %s; Value: %s; Updated at %s" % (self.index, self.value, self.updateTime)
        
    def _parseValue(self, value, dataType):
        """
        Convert an encoded value to something Pythonic (if possible).
        
        Understood codes:
         * NUL:   No data stored (e.g., branch head entry)
         * a####: printable (i.e., ASCII minus escape codes), #### = number of characters
                  e.g., "a3" means 3 printable ASCII-encoded characters
         * r####: raw data (not printable), #### = number of bytes
                  e.g., "r1024" means 1024 bytes of raw data
         * i1u:   integer, 1 byte,  unsigned, little-endian (=uint8)
         * i2u:   integer, 2 bytes, unsigned, litte-endian (=uint16)
         * i4u:   integer, 4 bytes, unsigned, little-endian (=uint32)
         * i4s:   integer, 4 bytes, signed, little-endian (=int32)
         * i8u:   integer, 8 bytes, unsigned, litte-endian (=uint64)
         * f4:    float, 4 bytes, little-endian (=float32)
         * f4r:   float, 4 bytes, big-ending (=float32)
        """
        
        if dataType == 'NUL':
            try:
                value = str(value, 'utf-8')
            except TypeError:
                value = str(value)
            return value
        elif dataType[0] == 'a':
            try:
                value = str(value, 'utf-8')
            except TypeError:
                value = str(value)
            return value
        elif dataType[0] == 'r':
            try:
                value = str(value, 'utf-8')
            except TypeError:
                value = str(value)
            return value

        elif dataType[:3] == 'i1u':
            try:
                return struct.unpack('<1B', value)[0]
            except struct.error:
                return 0
        elif dataType[:3] == 'i2u':
            try:
                return struct.unpack('<1H', value)[0]
            except struct.error:
                return 0
        elif dataType[:3] == 'i4u':
            try:
                return struct.unpack('<1I', value)[0]
            except struct.error:
                return 0
        elif dataType[:3] == 'i4s':
            try:
                return struct.unpack('<1i', value)[0]
            except struct.error:
                return 0
        elif dataType[:3] == 'i8u':
            try:
                return struct.unpack('<1Q', value)[0]
            except struct.error:
                return 0
        elif dataType[:3] == 'f4r':
            try:
                return struct.unpack('>1f', value)[0]
            except struct.error:
                return 0.0
        elif dataType[:2] == 'f4':
            try:
                return struct.unpack('<1f', value)[0]
            except struct.error:
                return 0.0
        else:
            raise ValueError("Unknown data type '%s'" % dataType)
            
    def fromEntry(self, value):
        """
        Given an MIB entry straight out of a GDBM database, populate the 
        MIBEntry instance.
        
        .. note::
            The MIBEntry class does not currently support branch entries.  
            Entries of this type will raise a ValueError.
            
        .. note::
            The MIBEntry class currently only support entries with indicies
            that start with a number.  All other indicies will raise a 
            ValueError.
        """
        
        # Initialize a new structure to parse the binary string
        record = parseCStruct("""
            int eType;
            char index[%i];
            char val[%i];
            char type_dbm[6];
            char type_icd[6];
            long tv[2];
            """ % (MIB_INDEX_FIELD_LENGTH, MIB_VAL_FIELD_LENGTH), endianness='little')
            
        # Squeeze the binary string into the structure using ctypes magic
        temp = ctypes.create_string_buffer(value)
        ctypes.memmove(ctypes.addressof(record), ctypes.addressof(temp), ctypes.sizeof(record))
        
        # Bytes to Unicode for Python3
        try:
            index   = str(record.index, 'utf-8')
            dbmType = str(record.type_dbm, 'utf-8')
            icdType = str(record.type_icd, 'utf-8')
        except Exception as e:
            index   = str(record.index)
            dbmType = str(record.type_dbm)
            icdType = str(record.type_icd)
            
        # Validate
        if record.eType == MIB_REC_TYPE_BRANCH:
            raise ValueError("Cannot interpret MIB branch entries")
        try:
            if index[0] not in ('0', '1', '2', '3', '4', '5', '6', '7', '8', '9'):
                raise ValueError("Entry index '%s' does not appear to be numeric" % record.index)
        except IndexError:
            raise ValueError("Entry index '%s' does not appear to be numeric" % record.index)
            
        # Basic information
        self.eType = int(record.eType)
        self.index = index
        self.value = self._parseValue(record.val, dbmType)
        self.dbmType = dbmType
        self.icdType = icdType
        self._tv = (int(record.tv[0]), int(record.tv[1]))
        
        # Time
        self.updateTime = datetime.utcfromtimestamp(record.tv[0] + record.tv[1]/1e9)
        
        return True



# -*- coding: utf-8 -*-

# Python3 compatiability
from __future__ import print_function, division
import sys
if sys.version_info > (3,):
    from functools import cmp_to_key

"""
Buffer for dealing with out-of-order/missing frames.

.. versionchanged:: 0.5
    Removed support for DRX FrameBuffers since they shouldn't be needed.
    
.. versionchanged:: 0.6
    Removed support for TBW FrameBuffers since they didn't really work.
    
.. versionchanged:: 1.0.0
    Added back in the DRX FrameBuffers since they might be useful.
    Dropped TBW support from within the FrameBuffer class.
    
.. versionchanged:: 1.0.2
    Added support for the LWA-SV ADP DRX8 mode
    
.. versionchanged:: 1.1.4
    Dropped support for the LWA-SV ADP DRX8 mode
    
.. versionchanged:: 1.2.1
    Added support for the LWA-SV TBF and COR modes
"""

import copy
from collections import deque
try:
    from collections import OrderedDict
except ImportError:
    from lsl.misc.OrderedDict import OrderedDict


__version__ = '1.2'
__revision__ = '$Rev: 2709 $'
__all__ = ['FrameBuffer', 'TBNFrameBuffer', 'DRXFrameBuffer', 'TBFFrameBuffer', 'VDIFFrameBuffer', '__version__', '__revision__', '__all__']


def _cmpFrames(x, y):
    """
    Function to compare two frames and sort by stand/beam number (TBN, DRX)
    or by channel number (TBF).
    """
    
    # Parse if frame IDs to extract the stand/beam, tunning, and polarization
    # information (where appropriate)
    try:
        idsX = x.parseID()
        idsY = y.parseID()
    except AttributeError:
        idsX = x.header.firstChan
        idsY = y.header.firstChan
        
    # Do a try...except block to catch TBW vs. TBN/DRX
    try:
        len1 = len(idsX)
        if len1 == 2:
            sX = 2*(idsX[0]-1) + idsX[1]
            sY = 2*(idsY[0]-1) + idsY[1]
        else:
            sX = 16777216*(idsX[0]-1) + 4096*(idsX[1]-1) + idsX[2]
            sY = 16777216*(idsY[0]-1) + 4096*(idsY[1]-1) + idsY[2]
    except TypeError:
        sX = idsX
        sY = idsY
        
    # Do the comparison
    if sY > sX:
        return -1
    elif sX > sY:
        return 1
    else:
        return 0
    

class FrameBuffer(object):
    """
    Frame buffer for re-ordering TBN and DRX frames in time order.  
    This class is filled with frames and a returns a frame list when 
    the 'nSegments' starts filling.  In that case, the oldest segment 
    is returned.

    The buffer also keeps track of what has already been read out so 
    that tardy frames are just dropped.  For buffers that are read out,
    missing frames are replaced with frames filled with zeros.
    
    .. note::
        Due to the nature of the buffer, it is possible that there will
        still be 'nSegments'-1 segements in the buffer that are either
        full or partially full.  This can be retrieved using the buffer's 
        'flush()' function.
    """
    
    def __init__(self, mode='TBN', stands=[], beams=[], tunes=[], pols=[], chans=[], threads=[], nSegments=6, ReorderFrames=False):
        """
        Initialize the buffer with a list of:
          * TBN
            * list of stands
            * list of pols
          * DRX
            * list of beams
            * list of tunings
            * list of pols.
          * TBF
            * list of start channels
          * COR
            * list of stands
            * list of start channels
          * VDIF
            * list of thread IDs
        By doing this, we should be able to keep up with when the buffer 
        is full and to help figure out which stands/beam/tunings/pols. are 
        missing.
        """
        
        # Input validation
        if mode.upper() not in ('TBN', 'DRX', 'TBF', 'COR', 'VDIF'):
            raise RuntimeError("Invalid observing mode '%s'" % mode)
            
        if mode.upper() == 'TBN':
            for pol in pols:
                if pol not in (0, 1):
                    raise RuntimeError("Invalid polarization '%i'" % pol)
                    
        elif mode.upper() == 'DRX':
            for tune in tunes:
                if tune not in (1, 2):
                    raise RuntimeError("Invalid tuning '%i'" % tune)
            for pol in pols:
                if pol not in (0, 1):
                    raise RuntimeError("Invalid polarization '%i'" % pol)
                    
        elif mode.upper() == 'TBF':
            for chan in chans:
                if chan not in range(4096):
                    raise RuntimeError("Invalid start channel '%i'" % chan)
                    
        elif mode.upper() == 'COR':
            for chan in chans:
                if chan not in range(4096):
                    raise RuntimeError("Invalid start channel '%i'" % chan)
            
        else:
            for thread in threads:
                if thread not in range(1024):
                    raise RuntimeError("Invalid thread ID '%i'" % thread)
                    
        # The buffer itself
        self.nSegments = nSegments
        self.buffer = OrderedDict()
        self.done = deque([0,], maxlen=self.nSegments)
        
        # Buffer statistics
        self.full = 0		# Number of times a full buffer was emptied
        self.partial = 0	# Number of times a partial buffer was emptied
        self.dropped = 0	# Number of late or duplicate frames dropped
        self.missing = 0	# Number of missing frames
        
        # Information describing the observations
        self.mode = mode.upper()
        self.stands = stands
        self.beams = beams
        self.tunes = tunes
        self.pols = pols
        self.chans = chans
        self.threads = threads
        
        # If we should reorder the returned frames by stand/pol or not
        self.reorder = ReorderFrames
        
        # Figure out how many frames fill the buffer and the list of all
        # possible frames in the data set
        self.nFrames, self.possibleFrames = self.calcFrames()
        self.possibleFrames = set(self.possibleFrames)
        
    @property
    def filled(self):
        """
        Indicates whether or not the ring buffer is full or not.
        
        .. versionadded:: 1.2.4
        """
        
        return False if len(self.buffer) < self.nSegments else True
        
    @property
    def overfilled(self):
        """
        Indicates whether or not the ring buffer has too many segements or not.
        
        .. versionadded:: 1.2.4
        """
        
        return True if len(self.buffer) > self.nSegments else False
        
    def calcFrames(self):
        """
        Calculate the maximum number of frames that we expect from 
        the setup of the observations and a list of tuples that describes
        all of the possible stand/pol combination.
        
        This will be overridden by sub-classes of FrameBuffer.
        """
        
        raise NotImplementedError
        
    def figureOfMerit(self, frame):
        """
        Figure of merit for storing/sorting frames in the ring buffer.
        
        This will be overridden by sub-classes of FrameBuffer.
        """
        
        raise NotImplementedError
        
    def createFill(self, key, frameParameters):
        """
        Create a 'fill' frame of zeros using an existing good
        packet as a template.
        
        This will be overridden by sub-classes of FrameBuffer.
        """
        
        raise NotImplementedError
        
    def append(self, frames):
        """
        Append a new frame to the buffer with the appropriate time tag.  
        True is returned if the frame was added to the buffer and False if 
        the frame was dropped because it belongs to a buffer that has 
        already been returned.
        """
        
        # Convert input to a deque (if needed)
        typeName = type(frames).__name__
        if typeName == 'deque':
            pass
        elif typeName == 'list':
            frames = deque(frames)
        else:
            frames = deque([frames,])
            
        # Loop over frames
        for frame in frames:
            # Make sure that it is not in the `done' list.  If it is,
            # disgaurd the frame and make a note of it.
            fom = self.figureOfMerit(frame)
            if fom <= self.done[-1]:
                self.dropped += 1
                continue
                
            # If that time tag hasn't been done yet, add it to the 
            # buffer in the correct place.
            try:
                self.buffer[fom].append(frame)
            except KeyError:
                self.buffer[fom] = deque()
                self.buffer[fom].append(frame)
                
        return True
        
    def put(self, *args, **kwds):
        """
        Synonymous with 'append'.
        """
        
        self.append(*args, **kwds)
        
    def peek(self):
        """
        Peek into the buffer to see what the next key to be retruned by a 
        call to get() will be.  Returns None if the buffer is not full.
        
        .. versionadded:: 1.2.4
        """
        
        # If the ring is not full, return nothing
        if not self.filled:
            return None
            
        # Get the current status of the buffer
        keys = list(self.buffer.keys())
        keyToReturn = min(keys)
        return keyToReturn
        
    def get(self, keyToReturn=None):
        """
        Return a list of frames that consitute a 'full' buffer.  Afterwards, 
        delete that buffer and mark it as closed so that any missing frames that
        are recieved late are dropped.  If none of the buffers are ready to be 
        dumped, None is returned.
        """
        
        if keyToReturn is None:
            keyToReturn = self.peek()
            if keyToReturn is None:
                return None
        returnCount = len(self.buffer[keyToReturn])
        
        if returnCount == self.nFrames:
            ## Everything is good (Is it really???)
            self.full = self.full + 1
            
            output = self.buffer[keyToReturn]
        elif returnCount < self.nFrames:
            ## There are too few frames
            self.partial = self.partial + 1
            self.missing = self.missing + (self.nFrames - returnCount)
            
            output = self.buffer[keyToReturn]
            
            ## Fill in the missing frames
            output.extend( map(lambda x: self.createFill(keyToReturn, x), self._missingList(keyToReturn)) )
        else:
            ## There are too many frames
            self.full = self.full + 1
            
            output = []
            frameIDs = []
            for frame in self.buffer[keyToReturn]:
                newID = frame.parseID()
                if newID not in frameIDs:
                    output.append(frame)
                    frameIDs.append(newID)
                else:
                    self.dropped += 1
                    
        del(self.buffer[keyToReturn])
        self.done.append(keyToReturn)
        
        # Sort and return
        if self.reorder:
            output = list(output)
            try:
                output.sort(cmp=_cmpFrames)
            except TypeError:
                output.sort(key=cmp_to_key(_cmpFrames))
        return output
        
    def flush(self):
        """
        Generator to return all of the remaining frames in the buffer 
        from buffers that are considered 'full'.  Afterwards, delete all 
        buffers.  This is useful for emptying the buffer after reading 
        in all of the data.
        
        .. note::
            It is possible for this function to return list of packets that
            are mostly invalid.
            
        .. versionchanged::1.2.1
            Converted to a generator to make sure that we can get all of 
            the LWA-SV COR frames out of the buffer
        """
        
        for key in self.buffer:
            yield self.get(keyToReturn=key)
            
    def reset(self):
        """
        Emtpy the contents of the buffer and reset it to a clean state.
        
        .. versionadded::1.2.1
        """
        
        self.buffer.clear()
        self.done.clear()
        self.done.append(0)
        
    def isEmpty(self):
        """
        Determine if there is anything in the buffer or not.  Returns False 
        if there is, True if there is not.
        
        .. versionadded::1.2.1
        """
        
        return False if len(self.buffer) else True
        
    def _missingList(self, key):
        """
        Create a list of tuples of missing frame information.
        """
        
        # Find out what frames we have
        if hasattr(self, 'frameID'):
            fnc = lambda x: self.frameID(x)
        else:
            if self.mode == 'VDIF':
                fnc = lambda x: x.parseID()[1]
            elif self.mode == 'TBF':
                fnc = lambda x: x.header.firstChan
            elif self.mode == 'COR':
                fnc = lambda x: x.parseID()+(x.header.firstChan,)
            else:
                fnc = lambda x: x.parseID()
        frameList = set(map(fnc, self.buffer[key]))
        
        # Compare the existing list with the possible list stored in the 
        # FrameBuffer object to build a list of missing frames.
        missingList = self.possibleFrames.difference(frameList)
        
        return missingList
        
    def status(self):
        """
        Print out the status of the buffer.  This contains information about:
          1. The current buffer fill level
          2. The numer of full and partial buffer dumps preformed
          3. The number of missing frames that fill packets needed to be created
             for
          4. The number of frames that arrived too late to be incorporated into 
             one of the ring buffers
        """
        
        nf = 0
        for key in self.buffer:
            nf = nf + len(self.buffer[key])
            
        outString = ''
        outString = '\n'.join([outString, "Current buffer level:  %i frames" % nf])
        outString = '\n'.join([outString, "Buffer dumps:  %i full / %i partial" % (self.full, self.partial)])
        outString = '\n'.join([outString, "--"])
        outString = '\n'.join([outString, "Missing frames:  %i" % self.missing])
        outString = '\n'.join([outString, "Dropped frames:  %i" % self.dropped])
        
        print(outString)


class TBNFrameBuffer(FrameBuffer):
    """
    A sub-type of FrameBuffer specifically for dealing with TBN frames.
    See :class:`lsl.reader.buffer.FrameBuffer` for a description of how the 
    buffering is implemented.
    
    Keywords:
      stands
        list of stand to expect packets for
    
      pols
        list of polarizations to expect packets for
    
      nSegments
        number of ring segments to use for the buffer (default is 20)
    
      ReorderFrames
        whether or not to reorder frames returned by get() or flush() by 
        stand/polarization (default is False)
    
    The number of segements in the ring can be converted to a buffer time in 
    seconds:
    
    +----------+------------------------------------------------+
    |          |                TBN Filter Code                 |
    | Segments +------+------+------+------+------+------+------+
    |          |  1   |  2   |  3   |  4   |  5   |  6   |  7   |
    +----------+------+------+------+------+------+------+------+
    |    10    |  5.1 |  1.6 |  0.8 |  0.4 |  0.2 |  0.1 | 0.05 |
    +----------+------+------+------+------+------+------+------+
    |    20    | 10.2 |  3.3 |  1.6 |  0.8 |  0.4 |  0.2 | 0.10 |
    +----------+------+------+------+------+------+------+------+
    |    30    | 15.4 |  4.9 |  2.5 |  1.2 |  0.6 |  0.3 | 0.15 |
    +----------+------+------+------+------+------+------+------+
    |    40    | 20.5 |  6.6 |  3.3 |  1.6 |  0.8 |  0.4 | 0.20 |
    +----------+------+------+------+------+------+------+------+
    |    50    | 25.6 |  8.2 |  4.1 |  2.0 |  1.0 |  0.5 | 0.26 |
    +----------+------+------+------+------+------+------+------+
    |   100    | 51.2 | 16.4 |  8.2 |  4.1 |  2.0 |  1.0 | 0.51 |
    +----------+------+------+------+------+------+------+------+
    
    """
    
    def __init__(self, stands=[], pols=[0, 1], nSegments=20, ReorderFrames=False):
        super(TBNFrameBuffer, self).__init__(mode='TBN', stands=stands, pols=pols, nSegments=nSegments, ReorderFrames=ReorderFrames)
        
    def calcFrames(self):
        """
        Calculate the maximum number of frames that we expect from 
        the setup of the observations and a list of tuples that describes
        all of the possible stand/pol combination.
        """
        
        nFrames = 0
        frameList = []
        
        nFrames = len(self.stands)*len(self.pols)
        for stand in self.stands:
            for pol in self.pols:
                frameList.append((stand,pol))
                
        return (nFrames, frameList)
        
    def figureOfMerit(self, frame):
        """
        Figure of merit for sorting frames.  For TBN it is:
            <frame timetag in ticks>
        """
        
        return frame.data.timeTag
        
    def createFill(self, key, frameParameters):
        """
        Create a 'fill' frame of zeros using an existing good
        packet as a template.
        """

        # Get a template based on the first frame for the current buffer
        fillFrame = copy.deepcopy(self.buffer[key][0])
        
        # Get out the frame parameters and fix-up the header
        stand, pol = frameParameters
        fillFrame.header.tbnID = 2*(stand-1) + pol + 1
        
        # Zero the data for the fill packet
        fillFrame.data.iq *= 0
        
        # Invalidate the frame
        fillFrame.valid = False
        
        return fillFrame


class DRXFrameBuffer(FrameBuffer):
    """
    A sub-type of FrameBuffer specifically for dealing with DRX frames.
    See :class:`lsl.reader.buffer.FrameBuffer` for a description of how the 
    buffering is implemented.
    
    Keywords:
      beams
        list of beam to expect packets for
    
      tunes
        list of tunings to expect packets for
    
      pols
        list of polarizations to expect packets for
    
      nSegments
        number of ring segments to use for the buffer (default is 20)
    
      ReorderFrames
        whether or not to reorder frames returned by get() or flush() by 
        stand/polarization (default is False)
    
    The number of segements in the ring can be converted to a buffer time in 
    seconds:
    
    +----------+--------------------------------------------------+
    |          |                 DRX Filter Code                  |
    | Segments +------+------+------+------+------+-------+-------+
    |          |  1   |  2   |  3   |  4   |  5   |  6    |  7    |
    +----------+------+------+------+------+------+-------+-------+
    |    10    | 0.16 | 0.08 | 0.04 | 0.02 | 0.01 | 0.004 | 0.002 |
    +----------+------+------+------+------+------+-------+-------+
    |    20    | 0.33 | 0.16 | 0.08 | 0.04 | 0.02 | 0.008 | 0.004 |
    +----------+------+------+------+------+------+-------+-------+
    |    30    | 0.49 | 0.25 | 0.12 | 0.06 | 0.03 | 0.013 | 0.006 |
    +----------+------+------+------+------+------+-------+-------+
    |    40    | 0.66 | 0.33 | 0.16 | 0.08 | 0.03 | 0.017 | 0.008 |
    +----------+------+------+------+------+------+-------+-------+
    |    50    | 0.82 | 0.41 | 0.20 | 0.10 | 0.04 | 0.021 | 0.010 |
    +----------+------+------+------+------+------+-------+-------+
    |   100    | 1.64 | 0.82 | 0.41 | 0.20 | 0.08 | 0.042 | 0.021 |
    +----------+------+------+------+------+------+-------+-------+
    
    """
    
    def __init__(self, beams=[], tunes=[1,2], pols=[0, 1], nSegments=20, ReorderFrames=False):
        super(DRXFrameBuffer, self).__init__(mode='DRX', beams=beams, tunes=tunes, pols=pols, nSegments=nSegments, ReorderFrames=ReorderFrames)
        
    def calcFrames(self):
        """
        Calculate the maximum number of frames that we expect from 
        the setup of the observations and a list of tuples that describes
        all of the possible stand/pol combination.
        """
        
        nFrames = 0
        frameList = []
        
        nFrames = len(self.beams)*len(self.tunes)*len(self.pols)
        for beam in self.beams:
            for tune in self.tunes:
                for pol in self.pols:
                    frameList.append((beam,tune,pol))
                    
        return (nFrames, frameList)
        
    def figureOfMerit(self, frame):
        """
        Figure of merit for sorting frames.  For DRX it is:
            <frame timetag in ticks>
        """
        
        return frame.data.timeTag
        
    def createFill(self, key, frameParameters):
        """
        Create a 'fill' frame of zeros using an existing good
        packet as a template.
        """

        # Get a template based on the first frame for the current buffer
        fillFrame = copy.deepcopy(self.buffer[key][0])
        
        # Get out the frame parameters and fix-up the header
        beam, tune, pol = frameParameters
        fillFrame.header.drxID = (beam & 7) | ((tune & 7) << 3) | ((pol & 1) << 7)
        
        # Zero the data for the fill packet
        fillFrame.data.iq *= 0
        
        # Invalidate the frame
        fillFrame.valid = False
        
        return fillFrame


class TBFFrameBuffer(FrameBuffer):
    """
    A sub-type of FrameBuffer specifically for dealing with TBF frames.
    See :class:`lsl.reader.buffer.FrameBuffer` for a description of how the 
    buffering is implemented.
    
    Keywords:
      chans
        list of start channel numbers to expect data for
    
      nSegments
        number of ring segments to use for the buffer (default is 25)
    
      ReorderFrames
        whether or not to reorder frames returned by get() or flush() by 
        start channel (default is False)
    
    The number of segements in the ring can be converted to a buffer time in 
    seconds:
    
    +----------+--------+
    | Segments |  Time  |
    +----------+--------+
    |    10    | 0.0004 |
    +----------+--------+
    |    25    | 0.001  |
    +----------+--------+
    |    50    | 0.002  |
    +----------+--------+
    |   100    | 0.004  |
    +----------+--------+
    |   200    | 0.008  |
    +----------+--------+
    
    """
    
    def __init__(self, chans, nSegments=25, ReorderFrames=False):
        super(TBFFrameBuffer, self).__init__(mode='TBF', chans=chans, nSegments=nSegments, ReorderFrames=ReorderFrames)
        
    def calcFrames(self):
        """
        Calculate the maximum number of frames that we expect from 
        the setup of the observations and a list of tuples that describes
        all of the possible stand/pol combination.
        """
        
        nFrames = 0
        frameList = []
        
        nFrames = len(self.chans)
        for chans in self.chans:
            frameList.append(chans)
            
        return (nFrames, frameList)
        
    def figureOfMerit(self, frame):
        """
        Figure of merit for sorting frames.  For TBF this is:
        frame.data.timeTag
        """
        
        return frame.data.timeTag
        
    def createFill(self, key, frameParameters):
        """
        Create a 'fill' frame of zeros using an existing good
        packet as a template.
        """

        # Get a template based on the first frame for the current buffer
        fillFrame = copy.deepcopy(self.buffer[key][0])
        
        # Get out the frame parameters and fix-up the header
        chan = frameParameters
        fillFrame.header.firstChan = chan
        
        # Zero the data for the fill packet
        fillFrame.data.fDomain *= 0
        
        # Invalidate the frame
        fillFrame.valid = False
        
        return fillFrame


class CORFrameBuffer(FrameBuffer):
    """
    A sub-type of FrameBuffer specifically for dealing with COR frames.
    See :class:`lsl.reader.buffer.FrameBuffer` for a description of how the 
    buffering is implemented.
    
    Keywords:
      chans
        list of start channel numbers to expect data for
    
      nSegments
        number of ring segments to use for the buffer (default is 5)
    
      ReorderFrames
        whether or not to reorder frames returned by get() or flush() by 
        start channel (default is False)
    
    The number of segements in the ring can be converted to a buffer time in 
    seconds:
    
    +----------+--------+
    | Segments |  Time  |
    +----------+--------+
    |     1    |   10   |
    +----------+--------+
    |     2    |   20   |
    +----------+--------+
    |     5    |   50   |
    +----------+--------+
    
    """
    
    def __init__(self, chans, nSegments=5, ReorderFrames=False):
        super(CORFrameBuffer, self).__init__(mode='COR', stands=list(range(1,256+1)), chans=chans, nSegments=nSegments, ReorderFrames=ReorderFrames)
        
    def calcFrames(self):
        """
        Calculate the maximum number of frames that we expect from 
        the setup of the observations and a list of tuples that describes
        all of the possible stand/pol combination.
        """
        
        nFrames = 0
        frameList = []
        
        nFrames = len(self.stands)*(len(self.stands)+1)//2 * len(self.chans)
        for stand0 in self.stands:
            for stand1 in self.stands:
                if stand1 < stand0:
                    continue
                for chan in self.chans:
                    frameList.append((stand0,stand1,chan))
                    
        return (nFrames, frameList)
        
    def figureOfMerit(self, frame):
        """
        Figure of merit for sorting frames.  For TBF this is:
        frame.data.timeTag
        """
        
        return frame.data.timeTag
        
    def createFill(self, key, frameParameters):
        """
        Create a 'fill' frame of zeros using an existing good
        packet as a template.
        """

        # Get a template based on the first frame for the current buffer
        fillFrame = copy.deepcopy(self.buffer[key][0])
        
        # Get out the frame parameters and fix-up the header
        stand0, stand1, chan = frameParameters
        fillFrame.header.firstChan = chan
        fillFrame.data.stand0 = stand0
        fillFrame.data.stand1 = stand1
        
        # Zero the data for the fill packet
        fillFrame.data.vis *= 0
        
        # Invalidate the frame
        fillFrame.valid = False
        
        return fillFrame


class VDIFFrameBuffer(FrameBuffer):
    """
    A sub-type of FrameBuffer specifically for dealing with VDIF frames.
    See :class:`lsl.reader.buffer.FrameBuffer` for a description of how the 
    buffering is implemented.
    
    Keywords:
      threads
        list of thread IDs to expect data for
    
      nSegments
        number of ring segments to use for the buffer (default is 10)
    
      ReorderFrames
        whether or not to reorder frames returned by get() or flush() by 
        stand/polarization (default is False)
    """
    
    def __init__(self, threads=[0,1], nSegments=10, ReorderFrames=False):
        super(VDIFFrameBuffer, self).__init__(mode='VDIF', threads=threads, nSegments=nSegments, ReorderFrames=ReorderFrames)
        
    def calcFrames(self):
        """
        Calculate the maximum number of frames that we expect from 
        the setup of the observations and a list of tuples that describes
        all of the possible stand/pol combination.
        """
        
        nFrames = 0
        frameList = []
        
        nFrames = len(self.threads)
        for thread in self.threads:
            frameList.append(thread)
            
        return (nFrames, frameList)
        
    def figureOfMerit(self, frame):
        """
        Figure of merit for sorting frames.  For VIDF this is:
        secondsFromEpoch * 100000 + frameInSecond
        """
        
        return frame.header.secondsFromEpoch*100000 + frame.header.frameInSecond
        
    def createFill(self, key, frameParameters):
        """
        Create a 'fill' frame of zeros using an existing good
        packet as a template.
        """

        # Get a template based on the first frame for the current buffer
        fillFrame = copy.deepcopy(self.buffer[key][0])
        
        # Get out the frame parameters and fix-up the header
        thread = frameParameters
        fillFrame.header.threadID = thread
        
        # Zero the data for the fill packet
        fillFrame.data.data *= 0
        
        # Invalidate the frame
        fillFrame.valid = False
        
        return fillFrame



# -*- coding: utf-8 -*-

# Python3 compatiability
import sys
if sys.version_info > (3,):
    xrange = range
    
"""Unit test for the lsl.misc.beamformer module."""

import os
import unittest
import numpy

from lsl.misc import beamformer
from lsl.common import stations


__revision__ = "$Rev: 2576 $"
__version__  = "0.2"
__author__    = "Jayce Dowell"


class beamformer_tests(unittest.TestCase):
    """A unittest.TestCase collection of unit tests for the lsl.sim.dp
    module."""
    
    def test_calcDelay(self):
        """Check that the beamformer.calcDelay function actually runs"""
        
        station = stations.lwa1
        antennas = station.getAntennas()
        
        out = beamformer.calcDelay(antennas[:3])
        self.assertEqual(len(out), 3)
        
        out = beamformer.calcDelay(antennas[:3], freq=49.0e6)
        self.assertEqual(len(out), 3)
        
        out = beamformer.calcDelay(antennas[:3], freq=49.0e6, azimuth=45, elevation=30)
        self.assertEqual(len(out), 3)
        
    def test_pointing_limits(self):
        """Test that beamformer.calcDelay respects the pointing limits"""
        
        station = stations.lwa1
        antennas = station.getAntennas()
        
        # Azimuth  checks
        self.assertRaises(beamformer.BeamformingError, beamformer.calcDelay, antennas[:3], 49.0e6, -5, 30)
        self.assertRaises(beamformer.BeamformingError, beamformer.calcDelay, antennas[:3], 49.0e6, 365, 30)
        
        # Elevation checks
        self.assertRaises(beamformer.BeamformingError, beamformer.calcDelay, antennas[:3], 49.0e6, 45, -5)
        self.assertRaises(beamformer.BeamformingError, beamformer.calcDelay, antennas[:3], 49.0e6, 45, 95)
        
    def test_intDelayAndSum(self):
        """Check that the beamformer.intDelayAndSum function actually runs"""
        
        station = stations.lwa1
        antennas = station.getAntennas()
        data = numpy.random.rand(3, 1000)
        
        beam = beamformer.intDelayAndSum(antennas[:3], data, azimuth=45.0, elevation=30.0)
        
    def test_intBeamShape(self):
        """Check that the beamformer.intBeamShape function actually runs"""
        
        station = stations.lwa1
        antennas = station.getAntennas()
        
        # Get a list of valid antennas
        ants = []
        for ant in antennas:
            if ant.pol == 1:
                continue
            if ant.getStatus() != 33:
                continue
                
            ants.append(ant)
            if len(ants) == 16:
                break
                
        # Multithreaded test for accuracy
        out = beamformer.intBeamShape(ants, azimuth=135.0, elevation=60.0)
        
        i = out.argmax()
        azDiff = numpy.abs(135.0 - i / 90)
        elDiff = numpy.abs(60.0 - i % 90)
        self.assertTrue(azDiff <= 1)
        self.assertTrue(elDiff <= 1)
        
        # Single threaded test for coverage
        out = beamformer.intBeamShape(ants[:1], azimuth=135.0, elevation=60.0, DisablePool=True)
            
    def test_phaseAndSum(self):
        """Check that the beamformer.phaseAndSum function actually runs"""
        
        station = stations.lwa1
        antennas = station.getAntennas()
        data = numpy.random.rand(3, 1000)
        
        beam = beamformer.phaseAndSum(antennas[:3], data, azimuth=45.0, elevation=30.0)
        
    def test_phaseBeamShape(self):
        """Check that the beamformer.phaseBeamShape function actually runs"""
        
        station = stations.lwa1
        antennas = station.getAntennas()
        
        # Get a list of valid antennas
        ants = []
        for ant in antennas:
            if ant.pol == 1:
                continue
            if ant.getStatus() != 33:
                continue
                
            ants.append(ant)
            if len(ants) == 16:
                break
                
        # Test for accuracy
        out = beamformer.phaseBeamShape(ants, azimuth=135.0, elevation=60.0)
        
        i = out.argmax()
        azDiff = numpy.abs(135.0 - i / 90)
        elDiff = numpy.abs(60.0 - i % 90)
        self.assertTrue(azDiff <= 1)
        self.assertTrue(elDiff <= 1)


class  beamformer_test_suite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the lsl.sim.vis units 
    tests."""
    
    def __init__(self):
        unittest.TestSuite.__init__(self)
        
        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(beamformer_tests)) 


if __name__ == '__main__':
    unittest.main()
